# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This file is used to define the basic graph."""
from collections import deque
from copy import deepcopy

from mindinsight.datavisual.data_transform.graph.msgraph import MSGraph
from mindinsight.debugger.common.exceptions.exceptions import \
    DebuggerNodeNotInGraphError
from mindinsight.debugger.common.log import LOGGER as log
from .node_type_identifier import NodeTypeIdentifier


def _is_match(identifier, node, condition):
    """Check if the node is matched to the identifier.
    Args:
        identifier (NodeTypeIdentifier): The debug name of the node.
        node (Node obj): The number of layers the user wants to trace. Default is 0.

    Returns:
        list, a list of the traced tensors' name and node type,
            arranged in order from leaf node to root node.
        int, the number of output tensors.
    """
    if condition:
        matched = identifier.is_match(node, condition)
    else:
        matched = identifier.is_match(node)
    return matched


class DebuggerGraph(MSGraph):
    """The `DebuggerGraph` object provides interfaces to describe a debugger graph."""

    def __init__(self):
        super(DebuggerGraph, self).__init__()
        self._root_graph_id = None
        self._is_root_graph = True
        self._graph_id = None

    def _parse_data(self, proto_data):
        super(DebuggerGraph, self)._parse_data(proto_data)
        root_name = proto_data.root_name
        graph_id_str = proto_data.name.split('_')[-1]
        self._is_root_graph = graph_id_str == root_name
        self._root_graph_id = int(root_name) if root_name.isdigit() else 0

    @property
    def root_graph_id(self):
        """Get the root graph id of current graph."""
        return self._root_graph_id

    @property
    def is_root_graph(self):
        """Return whether the graph is root graph."""
        return self._is_root_graph

    @property
    def leaf_nodes(self):
        """Return the leaf nodes."""
        return self._leaf_nodes

    @property
    def normal_node_map(self):
        """Return the normal_node_map"""
        return self._normal_node_map

    @property
    def node_id_map_name(self):
        """Return the node_id_map_name"""
        return self._node_id_map_name

    @property
    def const_node_temp_cache(self):
        """Return const_node_temp_cache"""
        return self._const_node_temp_cache

    @property
    def parameter_node_temp_cache(self):
        """Return parameter_node_temp_cache"""
        return self._parameter_node_temp_cache

    @property
    def full_name_map_name(self):
        """Return full_name_map_name"""
        return self._full_name_map_name

    def get_node_name_by_full_name(self, full_name):
        """Get node name by full names."""
        inner_name = self._full_name_map_name.get(full_name, '')
        if not inner_name:
            log.warning("Node %s does not find the relative inner node name.", full_name)

        return inner_name

    def get_full_name_by_node_name(self, node_name):
        """Get full name by node name."""
        if not node_name:
            return ''
        node = self._normal_node_map.get(node_name)
        if not node:
            log.error("Node <%s> is not in graph.", node_name)
            raise DebuggerNodeNotInGraphError(node_name=node_name)

        return node.full_name

    def get_node_type(self, node_name):
        """
        Get the type of the node.

        Args:
            node_name (str): The full name of the node with its scope.

        Returns:
            str, node type or name_scope.
        """
        if not node_name:
            return 'name_scope'
        node = self._normal_node_map.get(node_name)
        if not node:
            log.error("Node <%s> is not in graph.", node_name)
            raise DebuggerNodeNotInGraphError(node_name=node_name)

        return node.type

    def search_nodes_by_category(self, node_category, condition=None):
        """
        Search nodes by type.

        Args:
            node_category (TargetTypeEnum): The node type supported in
                mindinsight.conditionmgr.condition.TargetTypeEnum.
            condition (dict): Search condition. Default: None.

                - activation_func (Union[str, list[str]): The target functions. Used when node_type
                  is TargetTypeEnum.ACTIVATION.
                - search_range (list[Node]): The list of nodes to be searched from.

        Returns:
            list[Node], list of nodes.
        """
        identifier = NodeTypeIdentifier(node_category.value)
        # get search range
        condition = {} if condition is None else condition
        search_range = condition.pop('search_range', None)
        if not search_range:
            search_range = self._leaf_nodes.values()
        # search match nodes
        matched_nodes = []
        for node in search_range:
            matched = _is_match(identifier, node, condition)
            if matched:
                matched_nodes.append(node)
        return matched_nodes

    def get_tensor_history(self, node_name, depth=0):
        """
        Get the tensor history of a specified node.

        Args:
            node_name (str): The debug name of the node.
            depth (int): The number of layers the user wants to trace. Default is 0.

        Returns:
            list, a list of the traced tensors' name and node type,
                arranged in order from leaf node to root node.
            int, the number of output tensors.
        """
        node = self._leaf_nodes.get(node_name)
        tensor_history = self._get_tensor_infos_of_node(node)
        cur_outputs_nums = len(tensor_history)
        cur_depth = 0
        trace_list = deque([(node, cur_depth)])
        while trace_list:
            cur_node, cur_depth = trace_list.popleft()
            tensors_info = self._get_input_tensors_of_node(cur_node)
            if tensors_info:
                tensor_history.extend(tensors_info)
            if cur_depth < depth:
                for name in cur_node.inputs.keys():
                    trace_list.append((self._leaf_nodes[name], cur_depth + 1))

        return tensor_history, cur_outputs_nums

    @staticmethod
    def _get_tensor_infos_of_node(cur_node, slot=None):
        """Get tensors info of specified node."""
        tensors_info = []
        if slot is None:
            slots = range(cur_node.output_nums)
        elif slot >= 0:
            slots = [slot]
        else:
            log.info("Skip get tensor info for %s:%s.", cur_node.name, slot)
            return tensors_info
        for num in slots:
            tensor_info = {
                'name': cur_node.name + ':' + str(num),
                'full_name': cur_node.full_name + ':' + str(num),
                'node_type': cur_node.type
            }
            tensors_info.append(tensor_info)

        return tensors_info

    def _get_input_tensors_of_node(self, cur_node):
        """Get input tensors of node."""
        tensors_info = []
        for name in cur_node.inputs.keys():
            node = self._leaf_nodes.get(name)
            tensor_info = self._get_tensor_infos_of_node(node)
            tensors_info.extend(tensor_info)

        return tensors_info

    def get_tensor_graph(self, node_name):
        """
        Get graph relative to a node.

        Args:
            node_name (str): Node name.

        Returns:
            dict, tensor graph, format is:
                {'nodes': [
                    {'name': <node name>,
                    'full_name': <node full name>,
                    'type': <node type>
                    'input': <input objects>,
                    'output': <output objects>,
                    'slot': {'id': <slot id>},
                    'stack_info': [<stack info>]
                    }
                ]}
        """
        graph_nodes = []
        cur_node = self._leaf_nodes.get(node_name)
        node_detail_info = cur_node.to_dict()
        cur_node_info = self._get_node_info_for_tensor_graph(cur_node)
        cur_node_info['input'] = deepcopy(node_detail_info.get('input'))
        cur_node_info['output'] = deepcopy(node_detail_info.get('output'))
        self._add_input_node_info(cur_node_info=cur_node_info, graph_nodes=graph_nodes)
        self._add_output_node_info(cur_node=cur_node, cur_node_info=cur_node_info, graph_nodes=graph_nodes)
        graph_nodes.append(cur_node_info)
        return {'nodes': graph_nodes}

    @staticmethod
    def _get_node_info_for_tensor_graph(node):
        """Get node infos for tensor graph."""
        node_info = {
            'name': node.name,
            'full_name': node.full_name,
            'type': node.type,
            'input': {},
            'output': {},
            'slots': [{'slot': str(slot)} for slot in range(node.output_nums)],
            'stack_info': node.stack_info
        }
        return node_info

    def _add_output_node_info(self, cur_node, cur_node_info, graph_nodes):
        """
        Add output node info into cur_node_info and node list.

        Args:
            cur_node (Node): The current node object.
            cur_node_info (dict): Current node info.
            graph_nodes (list[<Node info>]): The nodes in tensor graph.
        """
        output_slot_mapping = self._get_slot_mapping(cur_node)
        for node_name, edge_info in cur_node_info.get('output').items():
            edge_info['slot_mapping'] = output_slot_mapping
            # add output node info into graph
            output_node = self._leaf_nodes.get(node_name)
            output_node_info = self._get_node_info_for_tensor_graph(output_node)
            output_node_info['input'][cur_node.name] = edge_info
            graph_nodes.append(output_node_info)

    def _add_input_node_info(self, cur_node_info, graph_nodes):
        """
        Add input node info into cur_node_info and node list.

        Args:
            cur_node_info (dict): Current node info.
            graph_nodes (list[<Node info>]): The nodes in tensor graph.
        """
        cur_node_name = cur_node_info.get('name')
        for node_name, edge_info in cur_node_info.get('input').items():
            input_node = self._leaf_nodes.get(node_name)
            edge_info['slot_mapping'] = self._get_slot_mapping(input_node)
            # add input node info into graph
            input_node_info = self._get_node_info_for_tensor_graph(input_node)
            input_node_info['output'][cur_node_name] = edge_info
            graph_nodes.append(input_node_info)

    @staticmethod
    def _get_slot_mapping(input_node):
        """Get slot mapping between nodes."""
        return [[str(slot), ''] for slot in range(input_node.output_nums)]



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The MindDataCpuUtilizationAnalyser analyser class."""
import csv
import json
import os

from mindinsight.profiler.analyser.base_analyser import BaseAnalyser
from mindinsight.profiler.common.exceptions.exceptions import ProfilerRawFileException, ProfilerFileNotFoundException
from mindinsight.profiler.common.log import logger as log
from mindinsight.profiler.common.validator.validate_path import validate_and_normalize_path
from mindinsight.profiler.analyser.minddata_analyser import MinddataAnalyser


class MinddataCpuUtilizationAnalyser(BaseAnalyser):
    """The analyser for analyzing minddata cpu utilization."""
    _cpu_utilization_display_filename = "minddata_cpu_utilization_{}.json"
    _minddata_pipeline_display_filename = "pipeline_profiling_{}.json"

    def __init__(self, profiling_dir, device_id):
        super(MinddataCpuUtilizationAnalyser, self).__init__(profiling_dir, device_id)
        self._step_total_num = 0
        self._steps_info = self._get_minddata_cpu_utilization_steps_info()
        self._cpu_utilization_info = dict()

    def get_idle_utilization_avg(self):
        """Get the idle utilization information of the whole machine."""
        filter_condition = {}
        self._filter(filter_condition)
        # When the idle utilization information is empty,
        # set the idle utilization rate to 100 to prevent the recommendation module from giving misleading suggestions
        # that the CPU utilization is too high.
        if not self._data.get("device_info").get("idle_utilization"):
            idle_utilization_avg = 100
            return idle_utilization_avg
        device_key_value = "device_info"
        self._get_cpu_utilization_average_value(device_key_value)
        idle_utilization_avg = self._cpu_utilization_info.get("device_info").get("idle_utilization").get("avg_value")
        return idle_utilization_avg

    def query(self, condition=None):
        """
        Query data according to the condition.

        Args:
            condition (dict): The search condition, only contains `filter_condition` parameter.
                Default: None.

        Returns:
            dict, the result after filtered, sorted and grouped.
        """
        if condition is None:
            condition = {}
        filter_condition = condition.get('filter_condition', {})
        log.info("Receive query request. %s", filter_condition)
        self._filter(filter_condition)
        self._cpu_utilization_info["sampling_interval"] = self._data.get("sampling_interval")
        self._cpu_utilization_info["step_info"] = self._steps_info
        self._cpu_utilization_info["step_total_num"] = self._step_total_num
        self._cpu_utilization_info["cpu_processor_num"] = self._data.get("cpu_processor_num")
        # device average CPU utilization
        device_key_value = "device_info"
        self._get_cpu_utilization_average_value(device_key_value)
        # process average CPU utilization
        process_key_value = "process_info"
        self._get_cpu_utilization_average_value(process_key_value)
        # op average CPU utilization
        self._get_cpu_utilization_op_average_value()
        return self._cpu_utilization_info

    def _load(self):
        """Load cpu_utilization info."""
        file_name = self._cpu_utilization_display_filename.format(self._device_id)
        file_path = os.path.join(self._profiling_dir, file_name)
        file_path = validate_and_normalize_path(
            file_path, raise_key="Invalid cpu_utilization_info file path.")
        if not os.path.exists(file_path):
            log.error('Did not find the cpu utilization file: %s', file_path)
            raise ProfilerFileNotFoundException(msg='Did not find the cpu utilization file.')

        with open(file_path, 'r', encoding='utf-8') as src_file:
            try:
                self._data = json.load(src_file)
            except json.JSONDecodeError as err:
                log.exception(err)
                raise ProfilerRawFileException("Fail to parse cpu_utilization info file")

    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.

                - start_step_id (int): The selected start step id.

                - end_step_id (int): The selected end step id.

        """
        start_step = filter_condition.get("start_step", 1)
        end_step = filter_condition.get("end_step", self._step_total_num)
        step_info_len = len(self._steps_info)
        if not step_info_len:
            return
        left_index = 0
        right_index = step_info_len - 1
        # Find the first point that greater than or equal to start_step.
        while left_index <= right_index:
            if int(self._steps_info[left_index]) >= start_step:
                break
            left_index += 1

        # Find the last point that smaller than or equal to end_step.
        while left_index <= right_index:
            if int(self._steps_info[right_index]) <= end_step:
                break
            right_index -= 1

        self._steps_info = self._steps_info[left_index:right_index + 1]
        # filter device CPU utilization
        for key in self._data.get("device_info").keys():
            self._data["device_info"][key] = \
                self._data.get("device_info").get(key)[left_index:right_index + 1]

        # filter process CPU utilization
        for key in self._data.get("process_info").keys():
            self._data["process_info"][key] = self._data.get("process_info").get(key)[left_index:right_index + 1]

        # filter op CPU utilization
        for item in self._data.get("op_info"):
            for key in item.get("metrics").keys():
                item["metrics"][key] = item.get("metrics").get(key)[left_index:right_index + 1]

    def _get_minddata_cpu_utilization_steps_info(self):
        """Establish a connection between cpu utilization sampling points and host queue capacity."""
        steps_info = []
        left_index = 0
        right_index = 0
        time_stamp = self._data.get("time_stamp")
        # Two sampling points are used to calculate the utilization,
        # and the second sampling point is used as the starting sampling point to calculate the step.
        time_stamp = time_stamp[1:]
        queue_step_time_info = self._get_minddata_queue_step_time_info()
        if queue_step_time_info:
            self._step_total_num = len(queue_step_time_info)
            step0 = 0
            for item in time_stamp:
                # queue_step_time_info[][0]:step_num
                # queue_step_time_info[][1]:sample time
                # points less than step1 are classified as step0
                if float(item) < float(queue_step_time_info[0][1]):
                    steps_info.append(step0)
                    continue
                while right_index < len(queue_step_time_info):
                    if float(item) <= float(queue_step_time_info[right_index][1]):
                        if float(item) < float(queue_step_time_info[right_index][1]):
                            steps_info.append(queue_step_time_info[left_index][0])
                        else:
                            steps_info.append(queue_step_time_info[right_index][0])
                        break
                    left_index = right_index
                    right_index += 1

                if right_index == len(queue_step_time_info):
                    steps_info.append(queue_step_time_info[right_index - 1][0])

            return steps_info
        self._get_step_total_num()
        return steps_info

    def _get_step_total_num(self):
        """Load data according to the parsed AICORE operator types file."""
        target_file = f'step_trace_raw_{self._device_id}_detail_time.csv'
        dynamic_shape_file = f'dynamic_shape_info_{self._device_id}.json'
        file_path = os.path.join(self._profiling_dir, target_file)
        dynamic_shape_file_path = os.path.join(self._profiling_dir, dynamic_shape_file)
        if dynamic_shape_file in os.listdir(self._profiling_dir):
            with open(dynamic_shape_file_path, 'r') as fp:
                shape_info = json.load(fp).get('op_type')
                for op_type in shape_info:
                    self._step_total_num = len(shape_info.get(op_type))
                    return
        if not file_path:
            log.error("Failed to find parsed trace time file.")
            raise ProfilerFileNotFoundException('parsed step trace time file.')
        file_path = validate_and_normalize_path(
            file_path, raise_key="Invalid latest_trace_trace_time file path.")
        with open(file_path, 'r') as handle:
            csv_reader = csv.reader(handle)
            data = list(csv_reader)
        self._step_total_num = len(data) - 2

    def _get_minddata_queue_step_time_info(self):
        """Get the sampling time information at the steps of the host queue"""
        minddata_queue_step_time_info = []
        minddata_analyser = MinddataAnalyser(self._profiling_dir, self._device_id)
        file_path = minddata_analyser.get_device_queue_file_path()
        if not os.path.exists(file_path):
            log.error('Did not find the device queue file: %s', file_path)
            return minddata_queue_step_time_info

        first_line = True
        expect_first_step_num = 1
        diff_value = 0
        with open(file_path) as data_file:
            for line in data_file.readlines():
                op_info = line.split()
                if first_line:
                    # if profiler is started in the process of training, the first step_num(op_info[2]) may not be 1.
                    # so we need convert it to start with 1.
                    diff_value = int(op_info[2]) - expect_first_step_num
                    first_line = False
                # op_info is a list like:['1','64','8','2','85406783']
                # The value of the first element in op_info is '0' or '1'.
                # '0' means that the time information is recorded.
                # '1' means that the queue information is recorded.
                # '1':queue info , '64':queue capacity, '8':step_num, '2':queue size, '85406783':sampling time.
                if op_info and op_info[0] == "1":
                    minddata_queue_step_time_info.append([str(int(op_info[2]) - diff_value), op_info[4]])
        return minddata_queue_step_time_info

    def _get_minddata_pipeline_info(self):
        """Get the number of thread cores in minddata pipeline operator"""
        file_name = self._minddata_pipeline_display_filename.format(self._device_id)
        file_path = os.path.join(self._profiling_dir, file_name)
        file_path = validate_and_normalize_path(
            file_path, raise_key="Invalid minddata_pipeline_info file path.")
        if not os.path.exists(file_path):
            log.error('Did not find the minddata_pipeline file: %s', file_path)
            raise ProfilerFileNotFoundException(msg='Did not find the minddata_pipeline file:{}'.format(file_path))

        with open(file_path, 'r', encoding='utf-8') as file:
            try:
                minddata_pipeline_info = json.load(file)
            except json.JSONDecodeError as err:
                log.exception(err)
                raise ProfilerRawFileException("Fail to parse minddata pipeline file")

        minddata_pipeline_op_info = []
        for item in minddata_pipeline_info.get("op_info"):
            op_info_dict = dict()
            op_info_dict["op_id"] = item.get("op_id")
            op_info_dict["num_workers"] = item.get("num_workers")
            minddata_pipeline_op_info.append(op_info_dict)
        return minddata_pipeline_op_info

    def _get_cpu_utilization_average_value(self, key_value):
        """Get cpu_utilization average value for host or process."""
        self._cpu_utilization_info[key_value] = dict()
        for key in self._data.get(key_value).keys():
            arr = self._data.get(key_value)[key]
            avg_value = round(sum(arr) / len(arr)) if arr else 0
            self._cpu_utilization_info[key_value][key] = {"metrics": arr, "avg_value": avg_value}

    def _get_cpu_utilization_op_average_value(self):
        """Get cpu_utilization average value for op."""
        minddata_pipeline_op_info = self._get_minddata_pipeline_info()
        self._cpu_utilization_info["op_info"] = {
            "op_list": [],
            "total_op_avg_value": {"user_utilization": 0, "sys_utilization": 0}
        }

        for item in self._data.get("op_info"):
            # Filtering out non minddata pipeline operator
            if str(item.get("op_id")) == "-1":
                continue
            op_info_dict = dict()
            op_info_dict["metrics"] = dict()
            for key in item.get("metrics").keys():
                arr = item.get("metrics")[key]
                avg_value = round(sum(arr) / len(arr)) if arr else 0
                op_info_dict["metrics"][key] = {"metrics": arr, "avg_value": avg_value}
                self._cpu_utilization_info["op_info"]["total_op_avg_value"][key] += avg_value
            op_info_dict["op_id"] = item.get("op_id")
            op_info = [i for i in minddata_pipeline_op_info if i.get("op_id") == item.get("op_id")]
            # op_info is like [{"num_workers":int,"op_id":int}]
            op_info_dict["num_workers"] = op_info[0].get("num_workers")
            self._cpu_utilization_info["op_info"]["op_list"].append(op_info_dict)



# Copyright 2020-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The gpu base analyser."""
import csv
import os
import json
from collections import defaultdict

from mindinsight.profiler.analyser.base_analyser import BaseAnalyser
from mindinsight.profiler.common.log import logger
from mindinsight.profiler.common.validator.validate_path import validate_and_normalize_path


class GpuAnalyser(BaseAnalyser):
    """Gpu base analyser."""
    _csv_file_to_analyse = ""

    def _load(self):
        """Load data according to the parsed AICORE operator types file."""
        op_type_file_path = os.path.join(
            self._profiling_dir,
            self._csv_file_to_analyse.format(self._device_id)
        )
        op_type_file_path = validate_and_normalize_path(
            op_type_file_path, raise_key="Invalid op_type_file_path")
        if not os.path.isfile(op_type_file_path):
            logger.warning('The file <%s> does not exist.', op_type_file_path)
            return

        with open(op_type_file_path, 'r') as file:
            csv_reader = csv.reader(file)
            _ = next(csv_reader)
            for info in csv_reader:
                self._data.append(self._convert_field_type(info))

    def _convert_field_type(self, row):
        """
        Convert the field type to the specific type.

        Args:
            row (list): One row data from parsed data.

        Returns:
            list, the converted data.
        """
        return row

    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.
        """

        def _inner_filter(item: list):
            return self._default_filter(item, filter_condition)

        self._result = list(filter(_inner_filter, self._data))


class GpuOpTypeAnalyser(GpuAnalyser):
    """Gpu operation type analyser."""
    _col_names = ["op_type", "total_time", "execution_frequency", "total_percent", "avg_time"]
    _csv_file_to_analyse = 'gpu_op_type_info_{}.csv'

    def _convert_field_type(self, row):
        """
        Convert the field type to the specific type.

        Args:
            row (list): One row data from parsed data.

        Returns:
            list, the converted data.
        """
        return [row[0], self._format_float_data(float(row[2])), int(row[1]),
                self._format_float_data(float(row[3]) * 100), self._format_float_data(float(row[4]))]


class GpuOpInfoAnalyser(GpuAnalyser):
    """Gpu operation detail info analyser."""
    _col_names = ["op_side", "op_type", "op_name", "op_full_name",
                  "op_occurrences", "op_total_time", "op_avg_time",
                  "proportion", "cuda_activity_cost_time", "cuda_activity_call_count"]
    _csv_file_to_analyse = 'gpu_op_detail_info_{}.csv'

    def _convert_field_type(self, row):
        """
        Convert the field type to the specific type.

        Args:
            row (list): One row data from parsed data.

        Returns:
            list, the converted data.
        """
        return [row[0], row[1], row[2], row[3], int(row[4]),
                self._format_float_data(float(row[5])), self._format_float_data(float(row[6])),
                self._format_float_data(float(row[7])), self._format_float_data(float(row[8])), int(row[9])]


class GpuCudaActivityAnalyser(GpuAnalyser):
    """Gpu activity type analyser."""
    _col_names = ["name", "type", "op_full_name", "stream_id",
                  "block_dim", "grid_dim", "occurrences", "total_duration",
                  "avg_duration", "max_duration", "min_duration"]
    _csv_file_to_analyse = 'gpu_activity_data_{}.csv'

    def _convert_field_type(self, row):
        """
        Convert the field type to the specific type.

        Args:
            row (list): One row data from parsed data.

        Returns:
            list, the converted data.
        """
        return [row[0], row[1], row[2], row[3], row[4], row[5], int(row[6]),
                self._format_float_data(float(row[7])), self._format_float_data(float(row[8])),
                self._format_float_data(float(row[9])), self._format_float_data(float(row[10]))]


class GpuDynamicAnalyser(BaseAnalyser):
    """Gpu dynamic shape data analyser."""
    _filter_type = defaultdict(list)
    _csv_file_to_analyse = 'dynamic_shape_info_{}.json'

    def query(self, condition=None):
        """
        Query data according to the condition.

        Args:
            condition (dict): The search condition, including filter condition,
                sort condition and group condition. If the condition is `None`,
                all data will be returned. Default: None.

        Returns:
            dict, the result after filtered, sorted and grouped.
        """
        if condition is None:
            condition = {}
        filter_condition = condition.get('filter_condition', {})
        sort_condition = condition.get('sort_condition')
        group_condition = condition.get('group_condition')

        self._result = []
        self._filter_type = defaultdict(list)
        self._display_col_names = self._col_names[:]
        self._filter(filter_condition)
        self._size = len(self._result)
        if sort_condition:
            self._sort(sort_condition)
        if group_condition:
            self._group(group_condition)
        return self._organize_query_result()

    def _load(self):
        """Load data according to the parsed dynamic shape file."""
        dynamic_shape_file_path = os.path.join(
            self._profiling_dir,
            self._csv_file_to_analyse.format(self._device_id)
        )
        dynamic_shape_file_path = validate_and_normalize_path(
            dynamic_shape_file_path, raise_key="Invalid dynamic shape file path.")
        if not os.path.isfile(dynamic_shape_file_path):
            logger.warning('The file <%s> does not exist.', dynamic_shape_file_path)
            return

        with open(dynamic_shape_file_path, 'r') as file:
            self._data = json.load(file)

    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.
        """
        default_type = "operator"
        limit = 3

        def _inner_filter(item: list):
            return self._default_filter(item, filter_condition)

        step = filter_condition.get("step_filter", ["1"])[0]
        step_info = self._data.get(self._switch_type, {}).get(step)
        if not step_info:
            return
        for item in step_info:
            self._result.append(list(item.values()))
        self._result = list(filter(_inner_filter, self._result))

        if self._switch_type == default_type:
            operator_data = self._data.get("operator_type")
        else:
            operator_data = self._data.get("kernel_type")
        self._all_type = list(operator_data[step].keys())

        if len(self._all_type) > limit:
            dispaly_op_type = filter_condition.get("dispaly_op_type", self._all_type[:limit])
        else:
            dispaly_op_type = filter_condition.get("dispaly_op_type", self._all_type)
        for step in operator_data:
            for op_type in dispaly_op_type:
                filter_data = operator_data[step].get(op_type, [0] * len(self._col_names))[2]
                self._filter_type[op_type].append(filter_data)

    def _organize_query_result(self):
        """
        Organize the query result.

        Returns:
            dict, the query result.
        """
        return {
            'col_name': self._display_col_names,
            'object': self._result,
            'size': self._size,
            "all_type": self._all_type,
            "filter_type": self._filter_type
        }


class GpuOpTypeInfoAnalyser(GpuDynamicAnalyser):
    """Gpu Op type info analyser."""
    _switch_type = "operator"
    _col_names = ["step", "op_side", "op_type", "op_name", "duration", "op_shape"]


class GpuCudaTypeInfoAnalyser(GpuDynamicAnalyser):
    """Gpu Cuda type info analyser."""
    _switch_type = "kernel"
    _col_names = ["step", "op_type", "op_name", "op_full_name", "duration", "block_dim", "grid_dim"]



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Modelarts exceptions module."""
from mindinsight.utils.exceptions import MindInsightException
from mindinsight.utils.constant import GeneralErrors


class PortReuseException(MindInsightException):
    """Port reuse exception."""
    def __init__(self, error_detail):
        error_msg = '{}'.format(error_detail)
        super(PortReuseException, self).__init__(
            GeneralErrors.PORT_NOT_AVAILABLE_ERROR,
            error_msg,
            http_code=400)



# Copyright 2020-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Validate the profiler parameters."""
import os
import sys

from mindinsight.datavisual.utils.tools import to_int
from mindinsight.profiler.common.exceptions.exceptions import ProfilerParamTypeErrorException, \
    ProfilerDeviceIdException, ProfilerOpTypeException, \
    ProfilerSortConditionException, ProfilerFilterConditionException, \
    ProfilerGroupConditionException, ProfilerParamValueErrorException
from mindinsight.profiler.common.log import logger as log

AICORE_TYPE_COL = ["op_type", "total_time", "execution_frequency", "percent"]
AICORE_DETAIL_COL = ["op_name", "op_type", "avg_execution_time", "execution_frequency", "MFLOPs(10^6 cube)",
                     "GFLOPS(10^9 cube)", "MFLOPs(10^6 vector)", "GFLOPS(10^9 vector)", "full_op_name",]
PYNATIVE_TYPE_COL = ["op_type", "execution_time", "execution_frequency", "percent"]
PYNATIVE_DETAIL_COL = ["op_name", "op_type", "avg_execution_time", "subgraph", "full_op_name"]
AICPU_TYPE_COL = ["op_type", "total_time", "execution_frequency", "percent"]
AICPU_DETAIL_COL = ["op_name", "serial_number", "op_type", "avg_execution_time", "dispatch_time", "execution_frequency",
                    "run_start", "run_end"]
GPU_TYPE_COL = ["op_type", "type_occurrences", "total_time", "proportion", "avg_time"]
GPU_OP_TYPE_COL = ["op_side", "op_type", "op_name", "duration"]
GPU_CUDA_TYPE_COL = ["op_type", "op_name", "op_full_name", "duration"]
GPU_ACTIVITY_COL = ["name", "type", "op_full_name", "stream_id",
                    "block_dim", "grid_dim", "occurrences", "total_duration",
                    "avg_duration", "max_duration", "min_duration"]
GPU_DETAIL_COL = ["op_side", "op_type", "op_name", "op_full_name",
                  "op_occurrences", "op_total_time", "op_avg_time",
                  "proportion", "cuda_activity_cost_time", "cuda_activity_call_count"]
CPU_TYPE_COL = ["op_type", "type_occurrences", "execution_frequency", "total_compute_time",
                "avg_time", "total_time_proportion"]
CPU_DETAIL_COL = ["op_side", "op_type", "op_name", "full_op_name", "op_occurrences",
                  "op_total_time", "op_avg_time", "total_time_proportion", "subgraph"]
MINDDATA_PIPELINE_COL = [
    'op_id', 'op_type', 'num_workers', 'output_queue_average_size',
    'output_queue_length', 'output_queue_usage_rate', 'sample_interval',
    'parent_id'
]


def validate_condition(search_condition):
    """
    Verify the param in search_condition is valid or not.

    Args:
        search_condition (dict): The search condition.

    Raises:
        ProfilerParamTypeErrorException: If the type of the param in search_condition is invalid.
        ProfilerDeviceIdException: If the device_id param in search_condition is invalid.
        ProfilerOpTypeException: If the op_type param in search_condition is invalid.
        ProfilerGroupConditionException: If the group_condition param in search_condition is invalid.
        ProfilerSortConditionException: If the sort_condition param in search_condition is invalid.
        ProfilerFilterConditionException: If the filter_condition param in search_condition is invalid.
    """
    if not isinstance(search_condition, dict):
        log.error("Invalid search_condition type, it should be dict.")
        raise ProfilerParamTypeErrorException(
            "Invalid search_condition type, it should be dict.")

    if "device_id" in search_condition:
        device_id = search_condition.get("device_id")
        if not isinstance(device_id, str):
            raise ProfilerDeviceIdException("Invalid device_id type, it should be str.")

    if "op_type" in search_condition:
        search_scope = get_search_scope(search_condition)
    else:
        raise ProfilerOpTypeException(
            "The op_type must in ['aicpu_type','aicpu_detail', 'aicore_type', 'aicore_detail', "
            "'gpu_op_type', 'gpu_op_info', 'gpu_cuda_activity', 'cpu_op_type', 'cpu_op_info', "
            "'gpu_op_type_info', 'gpu_cuda_type_info']")

    if "group_condition" in search_condition:
        validate_group_condition(search_condition)

    if "sort_condition" in search_condition:
        validate_sort_condition(search_condition, search_scope)

    if "filter_condition" in search_condition:
        validate_filter_condition(search_condition)


def get_search_scope(search_condition):
    """Get search scope."""
    op_type = search_condition.get("op_type")
    if op_type == "aicpu_type":
        search_scope = AICPU_TYPE_COL
    elif op_type == "aicpu_detail":
        search_scope = AICPU_DETAIL_COL
    elif op_type == "aicore_type":
        search_scope = AICORE_TYPE_COL
    elif op_type == "aicore_detail":
        search_scope = AICORE_DETAIL_COL
    elif op_type == "gpu_op_type":
        search_scope = GPU_TYPE_COL
    elif op_type == "gpu_op_info":
        search_scope = GPU_DETAIL_COL
    elif op_type == "gpu_cuda_activity":
        search_scope = GPU_ACTIVITY_COL
    elif op_type == "cpu_op_type":
        search_scope = CPU_TYPE_COL
    elif op_type == "cpu_op_info":
        search_scope = CPU_DETAIL_COL
    elif op_type == "pynative_type":
        search_scope = PYNATIVE_TYPE_COL
    elif op_type == "pynative_detail":
        search_scope = PYNATIVE_DETAIL_COL
    elif op_type == "gpu_op_type_info":
        search_scope = GPU_OP_TYPE_COL
    elif op_type == "gpu_cuda_type_info":
        search_scope = GPU_CUDA_TYPE_COL
    else:
        raise ProfilerOpTypeException(
            "The op_type must in ['aicpu_type','aicpu_detail', 'aicore_type', 'aicore_detail', "
            "'gpu_op_type', 'gpu_op_info', 'gpu_cuda_activity', 'cpu_op_type', 'cpu_op_info', "
            "'gpu_op_type_info', 'gpu_cuda_type_info']")
    return search_scope


def validate_group_condition(search_condition):
    """
    Verify the group_condition in search_condition is valid or not.

    Args:
        search_condition (dict): The search condition.

    Raises:
        ProfilerGroupConditionException: If the group_condition param in search_condition is invalid.
    """
    group_condition = search_condition.get("group_condition")
    if not isinstance(group_condition, dict):
        raise ProfilerGroupConditionException("The group condition must be dict.")
    if "limit" in group_condition:
        limit = group_condition.get("limit", 10)
        if isinstance(limit, bool) \
                or not isinstance(group_condition.get("limit"), int):
            log.error("The limit must be int.")
            raise ProfilerGroupConditionException("The limit must be int.")
        if limit < 1 or limit > 100:
            raise ProfilerGroupConditionException("The limit must in [1, 100].")

    if "offset" in group_condition:
        offset = group_condition.get("offset", 0)
        if isinstance(offset, bool) \
                or not isinstance(group_condition.get("offset"), int):
            log.error("The offset must be int.")
            raise ProfilerGroupConditionException("The offset must be int.")
        if offset < 0:
            raise ProfilerGroupConditionException("The offset must ge 0.")

        if offset > 1000000:
            raise ProfilerGroupConditionException("The offset must le 1000000.")


def validate_sort_condition(search_condition, search_scope):
    """
    Verify the sort_condition in search_condition is valid or not.

    Args:
        search_condition (dict): The search condition.
        search_scope (list): The search scope.

    Raises:
        ProfilerSortConditionException: If the sort_condition param in search_condition is invalid.
    """
    sort_condition = search_condition.get("sort_condition")
    if not isinstance(sort_condition, dict):
        raise ProfilerSortConditionException("The sort condition must be dict.")
    if "name" in sort_condition:
        sorted_name = sort_condition.get("name", "")
        err_msg = "The sorted_name must be in {}".format(search_scope)
        if not isinstance(sorted_name, str):
            log.error("Wrong sorted name type.")
            raise ProfilerSortConditionException("Wrong sorted name type.")
        if sorted_name not in search_scope:
            log.error(err_msg)
            raise ProfilerSortConditionException(err_msg)

    if "type" in sort_condition:
        sorted_type_param = ['ascending', 'descending']
        sorted_type = sort_condition.get("type")
        if sorted_type and sorted_type not in sorted_type_param:
            err_msg = "The sorted type must be ascending or descending."
            log.error(err_msg)
            raise ProfilerSortConditionException(err_msg)


def validate_op_filter_condition(op_condition, value_type=str, value_type_msg='str'):
    """
    Verify the op_condition in filter_condition is valid or not.

    Args:
        op_condition (dict): The op_condition in search_condition.
        value_type (type): The value type. Default: str.
        value_type_msg (str): The value type message. Default: 'str'.

    Raises:
        ProfilerFilterConditionException: If the filter_condition param in search_condition is invalid.
    """
    filter_key = ["in", "not_in", "partial_match_str_in", "display_op_type", "step"]
    if not isinstance(op_condition, dict):
        raise ProfilerFilterConditionException("The filter condition value must be dict.")
    for key, value in op_condition.items():
        if not isinstance(key, str):
            raise ProfilerFilterConditionException("The filter key must be str")
        if not isinstance(value, list):
            raise ProfilerFilterConditionException("The filter value must be list")
        if key not in filter_key:
            raise ProfilerFilterConditionException("The filter key must in {}.".format(filter_key))
        for item in value:
            if not isinstance(item, value_type):
                raise ProfilerFilterConditionException(
                    "The item in filter value must be {}.".format(value_type_msg)
                )


def validate_filter_condition(search_condition):
    """
    Verify the filter_condition in search_condition is valid or not.

    Args:
        search_condition (dict): The search condition.

    Raises:
        ProfilerFilterConditionException: If the filter_condition param in search_condition is invalid.
    """
    filter_condition = search_condition.get("filter_condition")
    if not isinstance(filter_condition, dict):
        raise ProfilerFilterConditionException("The filter condition must be dict.")
    if filter_condition:
        if "op_type" in filter_condition:
            op_type_condition = filter_condition.get("op_type")
            validate_op_filter_condition(op_type_condition)
        if "op_name" in filter_condition:
            op_name_condition = filter_condition.get("op_name")
            validate_op_filter_condition(op_name_condition)


def validate_and_set_job_id_env(job_id_env):
    """
    Validate the job id and set it in environment.

    Args:
        job_id_env (str): The id that to be set in environment parameter `JOB_ID`.

    Returns:
        int, the valid job id env.
    """
    if job_id_env is None:
        return job_id_env
    # get job_id_env in int type
    valid_id = to_int(job_id_env, 'job_id_env')
    # check the range of valid_id
    if valid_id and 255 < valid_id < sys.maxsize:
        os.environ['JOB_ID'] = job_id_env
    else:
        log.warning("Invalid job_id_env %s. The value should be int and between 255 and %s. Use"
                    "default job id env instead.",
                    job_id_env, sys.maxsize)
    return valid_id


def validate_ui_proc(proc_name):
    """
    Validate proc name in restful request.

    Args:
        proc_name (str): The proc name to query. Acceptable value is in
        [`iteration_interval`, `fp_and_bp`, `fp`, `tail`].

    Raises:
        ProfilerParamValueErrorException: If the proc_name is invalid.
    """
    accept_names = ['iteration_interval', 'fp_and_bp', 'fp', 'tail']
    if proc_name not in accept_names:
        log.error("Invalid proc_name. The proc_name for restful api is in %s", accept_names)
        raise ProfilerParamValueErrorException(f'proc_name should be in {accept_names}.')


def validate_minddata_pipeline_condition(condition):
    """
    Verify the minddata pipeline search condition is valid or not.

    Args:
        condition (dict): The minddata pipeline search condition.

    Raises:
        ProfilerParamTypeErrorException: If the type of the search condition is
            invalid.
        ProfilerDeviceIdException: If the device_id param in the search
            condition is invalid.
        ProfilerGroupConditionException: If the group_condition param in the
            search condition is invalid.
        ProfilerSortConditionException: If the sort_condition param in the
            search condition is invalid.
        ProfilerFilterConditionException: If the filter_condition param in the
            search condition is invalid.
    """
    if not isinstance(condition, dict):
        log.error("Invalid condition type, it should be dict.")
        raise ProfilerParamTypeErrorException(
            "Invalid condition type, it should be dict."
        )

    if "device_id" in condition:
        device_id = condition.get("device_id")
        if not isinstance(device_id, str):
            raise ProfilerDeviceIdException(
                "Invalid device_id type, it should be str."
            )

    if "group_condition" in condition:
        validate_group_condition(condition)

    if "sort_condition" in condition:
        validate_sort_condition(condition, MINDDATA_PIPELINE_COL)

    if "filter_condition" in condition:
        filter_condition = condition.get('filter_condition')
        if not isinstance(filter_condition, dict):
            raise ProfilerFilterConditionException(
                "The filter condition must be dict."
            )
        for key, value in filter_condition.items():
            if key == 'op_id':
                validate_op_filter_condition(
                    value, value_type=int, value_type_msg='int'
                )
            elif key == 'op_type':
                validate_op_filter_condition(value)
            elif key == 'is_display_op_detail':
                if not isinstance(value, bool):
                    raise ProfilerFilterConditionException(
                        "The condition must be bool."
                    )
            else:
                raise ProfilerFilterConditionException(
                    "The key {} of filter_condition is not support.".format(key)
                )



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Loss landscape processor Apis."""
from mindinsight.datavisual.common.enums import PluginNameEnum
from mindinsight.datavisual.common.exceptions import TrainJobNotExistError, LandscapeNotExistError
from mindinsight.datavisual.common.log import logger
from mindinsight.datavisual.common.validation import Validation
from mindinsight.datavisual.processors.base_processor import BaseProcessor
from mindinsight.lineagemgr.model import filter_summary_lineage
from mindinsight.utils.exceptions import MindInsightException, ParamValueError
from mindinsight.datavisual.data_transform.tensor_container import TensorContainer


class LandscapeProcessor(BaseProcessor):
    """Landscape processor."""

    def list_intervals(self, train_id):
        """Get the step intervals."""
        Validation.check_param_empty(train_id=train_id)
        train_job = self._data_manager.get_train_job_by_plugin(train_id, PluginNameEnum.LANDSCAPE.value)
        if train_job is None:
            raise TrainJobNotExistError("train_id {}".format(train_id))

        tags = train_job['tags']
        intervals = []
        for tag in tags:
            tensors = self._data_manager.list_tensors(train_id, tag)
            for tensor in tensors:
                loss_landscape = tensor.value
                if self._is_final_landscape(loss_landscape):
                    continue
                # Use the first value and last value to show the step or epoch interval
                intervals.append(dict(
                    id=self._get_interval_id(loss_landscape),
                    value=[loss_landscape.loss_path.intervals[0], loss_landscape.loss_path.intervals[-1]]
                ))
        intervals.sort(key=lambda interval: interval.get("value")[0])
        return dict(intervals=intervals)

    def list_landscapes(self, train_ids, landscape_type='interval', interval_id=None):
        """Get the landscape data for all train jobs."""
        landscapes = []
        for train_id in train_ids:
            if not isinstance(train_id, str) or not train_id.startswith('./'):
                logger.warning("The train id %s is invalid, it should be an relative path", train_id)
                landscape = {"error_code": ParamValueError().error_code, "train_id": train_id}
            else:
                landscape = self._get_landscape(train_id, landscape_type, interval_id)
            landscapes.append(landscape)
        return dict(landscapes=landscapes)

    def _get_landscape(self, train_id, landscape_type, interval_id):
        """Get the landscape data."""
        train_job = self._data_manager.get_train_job_by_plugin(train_id, PluginNameEnum.LANDSCAPE.value)
        if train_job is None:
            logger.warning("The train job for train id %s is not exist.", train_id)
            return {"error_code": TrainJobNotExistError().error_code, "train_id": train_id}

        for tag in train_job['tags']:
            landscape_events = self._data_manager.list_tensors(train_id, tag)
            for landscape_event in landscape_events:
                landscape = {}
                loss_landscape = landscape_event.value
                if landscape_type == 'interval' \
                        and not self._is_final_landscape(loss_landscape) \
                        and self._get_interval_id(loss_landscape) == interval_id:
                    landscape['train_id'] = train_id
                    landscape['points'] = self._get_points(loss_landscape)
                    landscape['path'] = self._get_path(loss_landscape)

                if landscape_type == 'final' and self._is_final_landscape(loss_landscape):
                    landscape['train_id'] = train_id
                    landscape['points'] = self._get_points(loss_landscape)

                if landscape:
                    landscape['metadata'] = self._get_metadata(loss_landscape, train_id)
                    landscape['convergence_point'] = self._get_convergence_point(loss_landscape)
                    return landscape

        logger.warning("The landscape value for train id %s is not found.", train_id)
        return {"error_code": LandscapeNotExistError().error_code, "train_id": train_id}

    @staticmethod
    def _get_interval_id(loss_landscape):
        """Get the interval id."""
        return str(hash(str(loss_landscape.loss_path.intervals)))

    @staticmethod
    def _is_final_landscape(loss_landscape):
        """If the landscape is the final one."""
        return not bool(loss_landscape.loss_path.intervals)

    def _get_points(self, loss_landscape):
        """Get points data."""
        points = dict(
            x=self._transform_tensor_to_list(loss_landscape.landscape.x),
            y=self._transform_tensor_to_list(loss_landscape.landscape.y),
            z=self._transform_tensor_to_list(loss_landscape.landscape.z)
        )
        return points

    def _get_path(self, loss_landscape):
        """Get the path of landscape."""
        path = dict(
            x=self._transform_tensor_to_list(loss_landscape.loss_path.points.x),
            y=self._transform_tensor_to_list(loss_landscape.loss_path.points.y),
            z=self._transform_tensor_to_list(loss_landscape.loss_path.points.z),
            intervals=list(loss_landscape.loss_path.intervals),
        )
        return path

    @staticmethod
    def _transform_tensor_to_list(tensor: TensorContainer):
        """Convert the tensor to list."""
        data = tensor.tensor_value
        return data.tolist()

    def _get_metadata(self, loss_landscape, train_id):
        """Get the metadata."""
        metadata = dict(
            decomposition=loss_landscape.metadata.decomposition,
            unit=loss_landscape.metadata.unit,
            step_per_epoch=loss_landscape.metadata.step_per_epoch
        )
        lineage_data = self._get_lineage_data(train_id)
        metadata.update(lineage_data)
        return metadata

    def _get_lineage_data(self, train_id):
        """Get the lineage data."""
        try:
            lineage_info = filter_summary_lineage(self._data_manager)
        except MindInsightException:
            lineage_info = {}
        lineage_objects = lineage_info.get("object", [])
        for lineage_data in lineage_objects:
            if lineage_data.get("summary_dir") == train_id:
                learning_rate = lineage_data.get('model_lineage', {}).get('learning_rate', None)
                loss = lineage_data.get('model_lineage', {}).get('loss', None)
                result = dict(
                    network=lineage_data.get('model_lineage', {}).get('network', None),
                    learning_rate=format(learning_rate, '6e') if learning_rate is not None else None,
                    optimizer=lineage_data.get('model_lineage', {}).get('optimizer', None),
                    metric=lineage_data.get('model_lineage', {}).get('metric', {}),
                    loss=format(loss, '6e') if loss is not None else None,
                )
                return result
        return dict(
            network=None,
            learning_rate=None,
            optimizer=None,
            metric={},
        )

    def _get_convergence_point(self, loss_landscape):
        """Get the convergence point data."""
        if loss_landscape.convergence_point.x.empty or loss_landscape.convergence_point.y.empty or \
                loss_landscape.convergence_point.z.empty:
            return [0, 0, 0]
        point = [
            self._transform_tensor_to_list(loss_landscape.convergence_point.x)[0],
            self._transform_tensor_to_list(loss_landscape.convergence_point.y)[0],
            self._transform_tensor_to_list(loss_landscape.convergence_point.z)[0],
        ]
        return point



# Copyright 2021-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Debugger Offline server."""
import copy
import os
import re
from collections import defaultdict
from importlib import import_module
from threading import Event

import mindinsight
from mindinsight.common.util import version_match
from mindinsight.conf import settings
from mindinsight.debugger.common.exceptions.exceptions import DebuggerModuleNotFoundError, DebuggerParamValueError, \
    DebuggerToolkitNotFoundError, DebuggerNodeTooLarge
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.common.utils import Streams, ServerStatus, DebuggerServerMode, get_ack_reply, \
    RunLevel, MAX_SINGLE_TENSOR_CACHE_BYTES, ViewCommandLevelEnum, convert_tensor_stats, put_tensor_base_in_cache, \
    put_tensor_stats_in_cache, get_tensor_value, MAX_MS_CACHE_SPACE_MB, get_download_file_name, add_to_download_mgr, \
    gc_disable, enter_tag
from mindinsight.debugger.conditionmgr.condition import ParamNameEnum
from mindinsight.debugger.debugger_services.debugger_server_base import DebuggerServerBase, debugger_server_wrap
from mindinsight.debugger.dump.convert import get_msaccucmp_path
from mindinsight.debugger.proto.debug_grpc_pb2 import EventReply
from mindinsight.debugger.stream_cache.data_loader import DataLoader
from mindinsight.debugger.stream_operator.watchpoint_operator import WatchpointOperator
from mindinsight.domain.graph.proto.ms_graph_pb2 import TensorProto
from mindinsight.utils.exceptions import MindInsightException


class DebuggerOfflineServer(DebuggerServerBase):
    """Debugger Offline Server."""
    _MAX_TRY_EXCEPT_COUNT = 500

    def __init__(self, cache_store, context):
        super(DebuggerOfflineServer, self).__init__(cache_store, context)
        self._offline_server_manager = DebuggerOfflineManager(cache_store, context.dbg_dir)
        self._running = Event()
        self._running.clear()

    def run(self):
        """Start the debugger offline server."""
        log.info("Initialize Offline Debugger Server for dbg_dir: %s", self._context.dbg_dir)
        self._offline_server_manager.initialize()
        log.info("Start Offline Debugger Server for dbg_dir: %s", self._context.dbg_dir)
        self._running.set()
        try_count = 0
        while self._running.is_set() and try_count < self._MAX_TRY_EXCEPT_COUNT:
            try:
                self._offline_server_manager.wait_for_termination()
                if not self._offline_server_manager.is_runnable():
                    break
            except MindInsightException as err:
                log.exception(err)
                log.warning("Error happens during listening on user commands. Restart listening again.")
            finally:
                try_count += 1
        # protect server from too much failure commands.
        if try_count == self._MAX_TRY_EXCEPT_COUNT:
            self._cache_store.clean()
            metadata = self._cache_store.get_stream_handler(Streams.METADATA).get()
            self._cache_store.put_data(metadata)
            log.warning("Exception exceed %d times, stop server.", try_count)

    def stop(self):
        """Stop offline debugger server."""
        if not self.is_alive():
            log.info("Offline debugger has already stop")
            return
        log.debug("Start to wait for thread started.")
        self._running.wait()
        log.info("Start to stop offline debugger server.")
        self._running.clear()
        self._offline_server_manager.stop()
        self.join()


class DebuggerOfflineManager:
    """Debugger offline manager which is used to handle user commands."""

    def __init__(self, cache_store, dbg_dir):
        cache_store.initialize()
        self._cache_store = cache_store
        self._metadata_stream = cache_store.get_stream_handler(Streams.METADATA)

        self._dbg_dir = dbg_dir
        self._dbg_services_module = self.get_dbg_service_module()
        self._dbg_service = None

        self._command_listener = CommandListener(cache_store)
        self._data_loader = DataLoader(dbg_dir)
        self.check_toolkit(self._data_loader)
        self._is_running_flag = False
        self._old_run_cmd = {}

    def stop(self):
        """Stop server."""
        self._is_running_flag = False
        self._command_listener.stop()
        self._cache_store.clean()
        event = get_ack_reply()
        event.exit = True
        self._cache_store.put_command(event)
        log.info("Stop debugger offline manager.")

    def is_runnable(self):
        """Check if the offline manager is runnable."""
        state = self._metadata_stream.state
        flag = self._is_running_flag and ServerStatus.is_normal_state(state)
        if not flag:
            log.debug("The offline manager is not runnable, is_running_flag: %s, metadata state: %s",
                      self._is_running_flag, state)
        return flag

    @staticmethod
    def get_dbg_service_module():
        """Get dbg service module from MindSpore."""
        try:
            dbg_services_module = import_module('mindspore.offline_debug.dbg_services')
        except (ModuleNotFoundError, ImportError) as err:
            log.error("Failed to find module dbg_services. %s", err)
            raise DebuggerModuleNotFoundError("dbg_services")
        return dbg_services_module

    @staticmethod
    def check_toolkit(data_loader):
        """Check if the toolkit file exist."""
        if not data_loader.get_sync_flag() and data_loader.device_target == "Ascend" and \
                data_loader.file_format == "bin":
            try:
                get_msaccucmp_path()
            except FileNotFoundError:
                msg = "Failed to find the msaccucmp.py or msaccucmp.pyc file. Please make sure you have installed " \
                      "toolkit package and set `ASCEND_TOOLKIT_PATH` in the environment correctly."
                log.error("File not found. %s", msg)
                raise DebuggerToolkitNotFoundError(msg)

    @staticmethod
    def get_dbg_service(dbg_services_module, data_loader, mem_limit):
        """Initialize DbgServices object."""
        is_sync = data_loader.get_sync_flag()
        net_name = data_loader.get_net_name()
        dump_dir = data_loader.get_dump_dir()
        dbg_service = dbg_services_module.DbgServices(dump_dir)
        try:
            dbg_service.initialize(net_name=net_name, is_sync_mode=is_sync, max_mem_usage=mem_limit)
        except TypeError:
            log.warning("The MindSpore and MindInsight version are mismatched. Failed to enable memory limit.")
            dbg_service.initialize(net_name=net_name, is_sync_mode=is_sync)
        return dbg_service

    @debugger_server_wrap
    def initialize(self):
        """Start to load offline debugger data."""
        self._dbg_service = self.get_dbg_service(self._dbg_services_module,
                                                 self._data_loader,
                                                 MAX_MS_CACHE_SPACE_MB)
        self._cache_store.clean()
        self._command_listener.start()
        self._is_running_flag = True
        self._check_version()
        if self._metadata_stream.state == ServerStatus.MISMATCH.value:
            log.info("The MindSpore and MindInsight version are mismatched. Failed to initialize offline server.")
            return
        self._load_metadata()
        self._load_graphs()
        log.info("Success initialize offline server for %s", self._dbg_dir)

    def _check_version(self):
        """Check version."""
        ms_version = self._dbg_services_module.get_version()
        mi_version = mindinsight.__version__
        self._metadata_stream.debugger_version = {'ms': ms_version, 'mi': mi_version}
        config_json = self._data_loader.get_config_json_data()
        ms_data_version = config_json.get("ms_version", None)
        data_match_stat = version_match(ms_data_version, mi_version)
        stat_reply = True
        if data_match_stat >= 1:
            stat_reply = False
        self._metadata_stream.data_version = {'state': stat_reply,
                                              'ms': ms_data_version, 'mi': mi_version}
        version_match_stat = version_match(ms_version, mi_version)
        if version_match_stat > 1:
            log.warning("Version is mismatched, dbg_services is: %s, mindinsight is: %s",
                        ms_version, mi_version)
            self._metadata_stream.state = ServerStatus.MISMATCH.value
            metadata = self._metadata_stream.get(['state', 'debugger_version', 'data_version'])
            self._cache_store.put_data(metadata)
        elif version_match_stat > 0:
            log.warning("Version is not completely matched, dbg_services is: %s, mindinsight is: %s",
                        ms_version, mi_version)

    def _load_metadata(self):
        """Load metadata."""
        self._metadata_stream.debugger_type = DebuggerServerMode.OFFLINE.value
        device_info = self._data_loader.load_device_info()
        # The backend referred to the running environment on which the offline debugger
        # data was generated.
        # Currently supported options: `GPU`, `Ascend`
        backend = device_info.get('device_target', 'Ascend')
        self._metadata_stream.backend = backend
        device_stream = self._cache_store.get_stream_handler(Streams.DEVICE)
        device_stream.put(device_info.get('server_list'))
        first_rank_info = device_stream.get()['devices'][0]
        self._metadata_stream.client_ip = first_rank_info.get('server_id')
        # get step number per device. dict(rank_id, step_num), may be increased with time goes by
        graph_history_stream = self._cache_store.get_stream_handler(Streams.GRAPH_HISTORY)
        graph_history_stream.put(self._data_loader.load_graph_history())
        graph_history_stream.put_dumped_step(self._data_loader.load_dumped_step())
        step_num_per_rank = graph_history_stream.get_total_count()
        device_stream.add_step_num_info(step_num_per_rank)
        self._metadata_stream.max_step_num = max(step_num_per_rank.values()) if step_num_per_rank else 0

    @gc_disable
    @enter_tag
    def _load_graphs(self):
        """Load graphs."""
        # the format of graphs is a list of {'rank_id': int, 'graph_protos': [GraphProto]}}
        try:
            graphs = self._data_loader.load_graphs(threshold=settings.MAX_GRAPH_NODE_SIZE)
        except DebuggerNodeTooLarge as err:
            self._metadata_stream.max_graph_node_size = settings.MAX_GRAPH_NODE_SIZE
            self._update_state(ServerStatus.NODE_TOO_LARGE)
            log.exception(err)
            return
        device_stream = self._cache_store.get_stream_handler(Streams.DEVICE)
        graph_per_rank = {}
        for graph in graphs:
            rank_id = graph.get('rank_id')
            graph_per_rank[rank_id] = {}
            tensor_stream_per_rank = self._cache_store.get_stream_handler(Streams.TENSOR). \
                get_tensor_handler_by_rank_id(rank_id, create_if_not_exit=True)
            for graph_proto in graph.get('graph_protos'):
                graph_per_rank[rank_id][graph_proto.name] = graph_proto
                tensor_stream_per_rank.put_const_vals(graph_proto.const_vals)
        # the graph_per_rank is format like: Dict[<rank_id>, Dict[<graph_name>, <GraphProto>]]
        try:
            self._cache_store.get_stream_handler(Streams.GRAPH).put(graph_per_rank)
            self._cache_store.get_stream_handler(Streams.GRAPH).parse_stack_infos()
            device_stream.add_graph_name_info(graph_per_rank)
        except DebuggerParamValueError:
            log.warning("Parse graph failed. The graph file is invalid.")
            self._cache_store.get_stream_handler(Streams.GRAPH).clean()
        self._metadata_stream.state = ServerStatus.RECEIVE_GRAPH.value

    @debugger_server_wrap
    def wait_for_termination(self):
        """Begin to listen on command event."""
        log.info("Begin to listen for user commands.")
        self._send_graph()
        while self.is_runnable():
            if not self._command_listener.has_new_command() and self._old_run_cmd:
                self._deal_with_old_run_cmd()
                continue
            cmd = self._command_listener.get_next_command()
            self.deal_with_cmd(cmd)

    def _send_graph(self):
        """Put graph and metadata info into data queue."""
        if not self.is_runnable():
            return
        self._metadata_stream.state = ServerStatus.WAITING.value
        metadata = self._metadata_stream.get()
        res = self._cache_store.get_stream_handler(Streams.GRAPH).get()
        res.update(metadata)
        self._cache_store.put_data(res)

    def _deal_with_old_run_cmd(self):
        """Deal with old run command."""
        left_step_count = self._old_run_cmd.get('left_step_count')
        if left_step_count:
            self._execute_one_step()
            # if old_run_cmd is not cleared due to hit.
            if self._old_run_cmd:
                self._old_run_cmd['left_step_count'] = left_step_count - 1 if left_step_count > 0 else -1
        if not self._old_run_cmd.get('left_step_count'):
            self._old_run_cmd.clear()

    def deal_with_cmd(self, cmd):
        """Deal with command."""
        if cmd is None:
            return
        if isinstance(cmd, dict):
            self._deal_with_view_cmd(cmd)
        elif isinstance(cmd, EventReply):
            self._on_event(cmd)

    def _on_event(self, event):
        """
        Deal with different command event.

        Args:
            event (EventReply): Command Event.
        """
        if event.HasField('run_cmd'):
            self._deal_with_run_cmd(event)
        elif event.HasField('exit'):
            self._cache_store.clean()
            self._update_state(ServerStatus.PENDING)
            log.debug("Clean cache for exit cmd.")
        else:
            self._deal_with_set_cmd(event)
            log.debug("Deal with set cmd.")

    def _deal_with_view_cmd(self, event):
        """
        Deal with view cmd.

        Args:
            event (dict): View command params.

                - view_cmd (EventReply): EventReply with view command.
                - node_name (str): The center node name for view command.
                - tensor_name (str): The center tensor name for view command.
                - graph_name (str): The graph name of center node.
                - rank_id (int): The device id of the tensor.
        """
        view_cmd = event.pop('view_cmd', None).view_cmd
        node_info = event
        log.debug("Receive view cmd for node: %s.", event)
        if not (view_cmd and node_info):
            log.info("Invalid view command. Ignore it.")
            return
        # read tensor value by dbg_service
        rank_id = node_info.get('rank_id', 0)
        cur_step = self._metadata_stream.step
        iteration_id = cur_step - 1
        root_graph_id = self.get_root_graph_id(rank_id=rank_id,
                                               graph_name=node_info.get('graph_name'),
                                               node_name=node_info.get('node_name'))
        tensor_protos = view_cmd.tensors
        tensor_infos = [
            self._dbg_services_module.TensorInfo(
                node_name=tensor_proto.node_name,
                slot=int(tensor_proto.slot),
                iteration=iteration_id - 1 if tensor_proto.iter == 'prev' else iteration_id,
                rank_id=rank_id,
                is_output=True,
                root_graph_id=root_graph_id
            ) for tensor_proto in tensor_protos]

        if cur_step <= 0:
            cst_infos = []
            cst_protos = []
            other_protos = []
            cst_prefix = "Default--data-"
            for tensor_proto in tensor_protos:
                if tensor_proto.node_name[0:len(cst_prefix)] == cst_prefix:
                    cst_infos.append(self._dbg_services_module.TensorInfo(
                        node_name=tensor_proto.node_name,
                        slot=int(tensor_proto.slot),
                        iteration=iteration_id - 1 if tensor_proto.iter == 'prev' else iteration_id,
                        rank_id=rank_id,
                        is_output=True,
                        root_graph_id=root_graph_id))
                    cst_protos.append(tensor_proto)
                else:
                    other_protos.append(tensor_proto)
            # if cur step is less than 1 and no cst tensor to read, put empty tensor value into cache.
            if not cst_protos:
                self._put_tensor_value_into_cache(cur_step, node_info, view_cmd.tensors)
                return
            self._put_tensor_value_into_cache(cur_step, node_info, other_protos)
            tensor_infos = cst_infos
            tensor_protos = cst_protos

        if node_info.get('level') == ViewCommandLevelEnum.BASE.value:
            self._read_tensor_base(cur_step, node_info, tensor_infos, tensor_protos)
        elif node_info.get('level') == ViewCommandLevelEnum.STATS.value:
            self._read_tensor_stats(cur_step, node_info, tensor_infos, tensor_protos)
        else:
            self._read_tensor_value(cur_step, node_info, tensor_infos, tensor_protos)

    def _read_tensor_base(self, cur_step, node_info, tensor_infos, tensor_protos):
        """Read tensor base from debugger service."""
        res = []
        tensor_data_res = self._dbg_service.read_tensor_base(tensor_infos)
        for tensor_data in tensor_data_res:
            tensor_data_dict = _convert_tensor_base(tensor_data)
            res.append(tensor_data_dict)
        rank_id = node_info.get('rank_id', 0)
        tensor_stream = self._cache_store.get_stream_handler(Streams.TENSOR). \
            get_tensor_handler_by_rank_id(rank_id)

        update_data_flag = put_tensor_base_in_cache(res, tensor_protos, cur_step, tensor_stream)

        if update_data_flag:
            self._send_update_msg(node_info)

    def _read_tensor_stats(self, cur_step, node_info, tensor_infos, tensor_protos):
        """Read tensor stats from debugger service."""
        res = []
        tensor_data_res = self._dbg_service.read_tensor_stats(tensor_infos)
        for tensor_data in tensor_data_res:
            tensor_data_dict = {
                'tensor_base': _convert_tensor_base(tensor_data),
                'tensor_stats': convert_tensor_stats(tensor_data)
            }
            res.append(tensor_data_dict)
        rank_id = node_info.get('rank_id', 0)
        tensor_stream = self._cache_store.get_stream_handler(Streams.TENSOR). \
            get_tensor_handler_by_rank_id(rank_id)

        update_data_flag = put_tensor_stats_in_cache(res, tensor_protos, cur_step, tensor_stream)

        if update_data_flag:
            self._send_update_msg(node_info)

    def _read_tensor_value(self, cur_step, node_info, tensor_infos, tensor_protos):
        """Read tensor value from debugger service."""
        res = []
        tensor_data_res = self._dbg_service.read_tensors(tensor_infos)
        for tensor_data in tensor_data_res:
            tensor_data_dict = _convert_tensor_value(tensor_data)
            res.append(tensor_data_dict)
        if len(tensor_protos) != len(res):
            log.error("Invalid view command. The result unmatched. %d/%d", len(tensor_protos), len(res))
        else:
            # put tensor into cache
            for tensor_proto, tensor_data in zip(tensor_protos, res):
                log.debug("Tensor name: %s:%s, tensor type: %s, tensor size: %s",
                          tensor_proto.node_name, tensor_proto.slot,
                          tensor_data.get('dtype'), tensor_data.get('data_size'))
                tensor_proto.tensor_content = tensor_data.get('data_ptr')
                tensor_proto.ClearField('dims')
                tensor_proto.dims.extend(tensor_data.get('shape'))
                tensor_proto.data_type = tensor_data.get('dtype')

        self._put_tensor_value_into_cache(cur_step, node_info, tensor_protos)
        log.info("Put tensor value into cache.")

    def get_root_graph_id(self, rank_id, graph_name, node_name=None):
        """Get root graph id."""
        graph_stream = self._cache_store.get_stream_handler(Streams.GRAPH).get_graph_handler_by_rank_id(rank_id)
        return graph_stream.get_root_graph_id(graph_name, node_name)

    def _put_tensor_value_into_cache(self, cur_step, node_info, tensor_protos):
        """Put tensor value into tensor cache."""
        rank_id = node_info.get('rank_id', 0)
        tensor_stream = self._cache_store.get_stream_handler(Streams.TENSOR). \
            get_tensor_handler_by_rank_id(rank_id)
        update_data_flag = False
        for tensor_proto in tensor_protos:
            if not tensor_proto.tensor_content:
                log.warning("Tensor %s:%s is empty.",
                            tensor_proto.node_name, tensor_proto.slot)
            try:
                load_info = node_info.get('load')
                if load_info is not None:
                    load_info['graph_name'] = node_info.get('graph_name')
                    load_info['node_name'] = node_info.get('node_name')
                    self._load_tensor(tensor_proto.node_name, tensor_proto.slot, cur_step, rank_id, load_info)
                oversize = len(tensor_proto.tensor_content) > MAX_SINGLE_TENSOR_CACHE_BYTES
                tensor_contents = [tensor_proto.tensor_content]
                data_size = len(tensor_proto.tensor_content)
                value = get_tensor_value(tensor_proto, tensor_contents, node_info, cur_step, oversize, data_size)
                has_update = tensor_stream.put(value)

            except ValueError as err:
                log.warning("Failed to put %s:%s into cache. Ignore it. %s",
                            tensor_proto.node_name, tensor_proto.slot, str(err))
                continue
            if has_update:
                update_data_flag = True
        if update_data_flag:
            self._send_update_msg(node_info)

    def _send_update_msg(self, node_info):
        # send message to frontend.
        metadata = self._metadata_stream.get(['step', 'state'])
        ret = {'receive_tensor': node_info.copy()}
        ret.update(metadata)
        self._cache_store.put_data(ret)

    def _deal_with_run_cmd(self, event):
        """Deal with run cmd."""
        run_cmd = event.run_cmd
        parsed_run_cmd = self._get_parsed_run_cmd(run_cmd)
        if parsed_run_cmd.run_steps > 0:
            self._execute_one_step()
        elif run_cmd.run_level == RunLevel.RECHECK.value:
            log.info("Deal with recheck command.")
            self._check_watchpoint(self._metadata_stream.step)

    def _execute_one_step(self):
        """Execute on step."""
        new_step = self._metadata_stream.step + 1
        if new_step > self._metadata_stream.max_step_num:
            self._old_run_cmd.clear()
            log.info("The server is already at the last step. %s", self._metadata_stream.max_step_num)
            return
        log.info("Go to next step: %s.", new_step)
        self._cache_store.clean_data()
        self._metadata_stream.step = new_step
        self._cache_store.get_stream_handler(Streams.TENSOR).set_step(new_step)
        self._check_watchpoint(new_step)
        self._cache_store.put_data(self._metadata_stream.get(['step', 'state']))

    def _get_parsed_run_cmd(self, run_cmd):
        """Get parsed run command."""
        if run_cmd.run_level == RunLevel.STEP.value:
            # receive pause cmd
            if not run_cmd.run_steps:
                log.debug("Pause training and wait for next command.")
                self._old_run_cmd.clear()
                # update metadata state from sending to waiting
                self._update_state(ServerStatus.WAITING)
                return run_cmd
            # receive step cmd
            left_steps = run_cmd.run_steps - 1
            run_cmd.run_steps = 1
            if left_steps:
                self._old_run_cmd['left_step_count'] = left_steps if left_steps > 0 else -1
        elif run_cmd.node_name:
            self._old_run_cmd['node_name'] = run_cmd.node_name
            run_cmd.node_name = ''
        return run_cmd

    def _check_watchpoint(self, step):
        """
        Save watchpoint hits into cache.

        Args:
            step (int): The step number which starts from 1.
        """
        self._update_state(ServerStatus.RUNNING)
        # Clean watchpoint_hits in cache
        multi_card_hit_streams = self._cache_store.get_stream_handler(Streams.WATCHPOINT_HIT)
        multi_card_hit_streams.clean()
        if step <= 0:
            log.info("Offline debugger does not support checking initial weights yet.")
            return
        if self._cache_store.get_stream_handler(Streams.WATCHPOINT).empty:
            log.debug("No watchpoint is set. Ignore checking watchpoint.")
            return
        # the iteration number in dump structure which starts from 0.
        iteration_id = step - 1
        if not self._data_loader.has_data(iteration_id):
            log.info("No data dumped with iteration id: %s. Ignore checking watchpoint.", iteration_id)
            return
        hits = []
        res = self._dbg_service.check_watchpoints(iteration_id)
        for watchpoint_hit in res:
            hit_dict = convert_watchpointhit(watchpoint_hit)
            hits.append(hit_dict)
        log.debug("Before sorting, unsorted_hits are: %s.", hits)
        # Sort the watchpointhits by timestamp
        hits.sort(key=lambda x: self.find_timestamp(x, iteration_id))
        log.debug("After sorting, hits are: %s.", hits)
        log.info("finish check watchpoint of %s", step)
        if hits:
            log.info("Received WatchpointHits. Left run cmd %s change to empty.", self._old_run_cmd)
            self._old_run_cmd.clear()
            self._update_state(ServerStatus.WAITING)
            self._save_watchpoint_hits(hits)

    def _save_watchpoint_hits(self, hits):
        """Save watchpoint hits."""
        multi_card_hit_streams = self._cache_store.get_stream_handler(Streams.WATCHPOINT_HIT)
        multi_card_graph_streams = self._cache_store.get_stream_handler(Streams.GRAPH)
        watchpoint_stream = self._cache_store.get_stream_handler(Streams.WATCHPOINT)

        watchpoint_hits = defaultdict(list)
        for hit in hits:
            log.info("Received hit\n: "
                     "name:%s, slot:%s, condition:%s, "
                     "watchpoint_id:%s"
                     "error_code:%s, rank_id:%s",
                     hit['name'], hit['slot'], hit['condition'],
                     hit['watchpoint_id'], hit['error_code'], hit['rank_id'])
            rank_id = hit['rank_id']
            watchpoint_hit = {}
            self._add_hit_node_info(watchpoint_hit, multi_card_graph_streams, rank_id, hit)
            if not watchpoint_hit:
                continue
            self._add_hit_watchpoint_info(watchpoint_hit, watchpoint_stream, hit)
            watchpoint_hit['error_code'] = hit['error_code']
            watchpoint_hits[rank_id].append(watchpoint_hit)
        # save hit info into cache
        multi_card_hit_streams.put(watchpoint_hits)
        self._cache_store.put_data({'receive_watchpoint_hits': True})
        log.debug("Send the watchpoint hits to DataQueue.")

    @staticmethod
    def _add_hit_node_info(watchpoint_hit, multi_card_graph_streams, rank_id, hit):
        """Add hit node info."""
        graph_stream = multi_card_graph_streams.get_graph_handler_by_rank_id(rank_id)
        node_full_name = hit['name']
        graph_name = graph_stream.get_graph_id_by_full_name(node_full_name)
        if not graph_name:
            log.warning("Cannot find node %s in graph. Skip it.", node_full_name)
            return
        ui_node_name = graph_stream.get_node_name_by_full_name(node_full_name, graph_name)
        log.debug("Receive watch point hit: %s:%s", node_full_name, hit['slot'])
        if not ui_node_name:
            log.info("Not support to show %s on graph.", node_full_name)
            return
        watchpoint_hit.update({
            'tensor_proto': TensorProto(node_name=node_full_name, slot=str(hit['slot'])),
            'node_name': ui_node_name,
            'graph_name': graph_name
        })

    @staticmethod
    def _add_hit_watchpoint_info(watchpoint_hit, watchpoint_stream, hit):
        """Add watchpoint hit info."""
        watchpoint = copy.deepcopy(watchpoint_stream.get_watchpoint_by_id(hit['watchpoint_id']))
        hit_params = {}
        # get hit actual value
        for param in hit['parameters']:
            if param['name'] not in (ParamNameEnum.RTOL.value, ParamNameEnum.RANGE_START_INCLUSIVE.value,
                                     ParamNameEnum.RANGE_END_INCLUSIVE.value) \
                    and hit['error_code'] == 0:
                hit_params[param['name']] = param['actual_value']
        # update actual value into watchpoint
        watchpoint_condition_params = watchpoint.condition['params']
        for i, param in enumerate(watchpoint_condition_params):
            name = param['name']
            if name in hit_params.keys():
                watchpoint_condition_params[i]['actual_value'] = hit_params[name]
            else:
                watchpoint_condition_params[i]['actual_value'] = None

        watchpoint_hit['watchpoint'] = watchpoint

    def _deal_with_set_cmd(self, event):
        """
        Deal with set cmd.

        Args:
            event (EventReply): User command event including set_cmd.
        """
        set_cmd = event.set_cmd
        set_cmd_id = set_cmd.id
        watchpoint_stream = self._cache_store.get_stream_handler(Streams.WATCHPOINT)
        watchpoint_stream.clean_cache_set_cmd(set_cmd)
        if not set_cmd.delete:
            log.info("Add watchpoint by using dbg_server.")
            multi_card_graph_stream = self._cache_store.get_stream_handler(Streams.GRAPH)
            WatchpointOperator.add_set_cmd_detail(set_cmd, watchpoint_stream, multi_card_graph_stream)
            watch_condition = set_cmd.watch_condition
            param_list = []
            for param in watch_condition.params:
                param_list.append(
                    self._dbg_services_module.Parameter(param.name, param.disabled, param.value))
            watch_nodes = set_cmd.watch_nodes
            check_nodes = self._get_check_nodes(watch_nodes)
            log.debug("Watchpoint  %s, condition: %s, watch nodes: %s",
                      set_cmd_id, watch_condition.condition, check_nodes)
            self._dbg_service.add_watchpoint(set_cmd_id, watch_condition.condition, check_nodes, param_list)
        else:
            log.info("Remove watchpoint by using dbg_server.")
            self._dbg_service.remove_watchpoint(set_cmd_id)

    def _get_check_nodes(self, watch_nodes):
        """Get check nodes format."""
        check_nodes = {}
        for watch_node in watch_nodes:
            node_name = watch_node.node_name
            rank_id = watch_node.rank_id
            root_graph_id = self.get_root_graph_id(rank_id=rank_id,
                                                   graph_name=watch_node.graph_name,
                                                   node_name=node_name)
            if node_name not in check_nodes:
                check_nodes[node_name] = {
                    "rank_id": set([rank_id]),
                    "is_output": True,
                    "root_graph_id": set([root_graph_id])
                }
            else:
                check_nodes[node_name]["rank_id"].add(rank_id)
                check_nodes[node_name]["root_graph_id"].add(root_graph_id)
        return check_nodes

    def _update_state(self, server_status):
        """
        Update state in metadata stream.

        Args:
            server_status (ServerStatus): The enum value in ServerStatus.
        """
        if self._metadata_stream.state != server_status.value:
            self._metadata_stream.state = server_status.value
            self._cache_store.clean_data()
            self._cache_store.put_data(self._metadata_stream.get())

    def find_timestamp(self, watchpointhit, step):
        """
        Return the time stamp of the watchpointhit.

        Args:
            watchpointhit (dict): The watchpointhit dict.

        Returns:
            int, the time stamp of the watchpointhit.
        """
        res = 0
        name = watchpointhit['name']
        slot = watchpointhit['slot']
        iteration = step
        rank_id = watchpointhit['rank_id']
        root_graph_id = watchpointhit['root_graph_id']
        tensor_files = self.find_tensor_file(name, slot, iteration, rank_id, root_graph_id)
        node_name = name.rsplit('/')[-1]
        for tensor_file in tensor_files:
            try:
                file_name = tensor_file
                node_name_location = file_name.find(node_name)
                post_str = file_name[node_name_location + len(node_name):]
                time_str = post_str.split('.')[3]
                time_stamp = int(time_str)
                if time_stamp > res:
                    res = time_stamp
            except (ValueError, TypeError, IndexError) as err:
                log.error("Can not find time_stamp in file: %s, the detail error is: %s.", tensor_file, err)
                continue
        log.debug("Time stamp is: %s.", res)
        return res

    def find_tensor_file(self, name, slot, iteration, rank_id, root_graph_id):
        """Find the tensor file."""
        iteration_path = self.get_iteration_path(iteration, root_graph_id, rank_id)
        # Find the pure node_name without scope
        node_name = name.rsplit('/')[-1]
        tensor_files = [fname for fname in os.listdir(iteration_path) if
                        re.match(r"[^\.]+\.{}\.[0-9]+\.[0-9]+\.[0-9]+\.output\.{}.*\.npy".format(node_name, slot),
                                 fname)]
        log.debug("Find %s files in path: %s.", len(tensor_files), iteration_path)
        return tensor_files

    def get_iteration_path(self, iteration, root_graph_id, rank_id):
        """Get the path of tensors in current iteration."""
        dump_dir = self._data_loader.get_dump_dir()
        rank_str = 'rank_' + str(rank_id)
        iteration_path = os.path.join(dump_dir, rank_str, self._data_loader.get_net_name(),
                                      str(root_graph_id), str(iteration))
        return iteration_path

    def _load_tensor(self, name, slot, step, rank_id, load_info):
        """Find and prepare the tensor."""
        if load_info.get('prev') == 'prev':
            step -= 1
        root_graph_id = self.get_root_graph_id(rank_id=rank_id, graph_name=load_info.get("graph_name"),
                                               node_name=load_info.get("node_name"))
        iteration_path = self.get_iteration_path(step - 1, root_graph_id, rank_id)
        tensor_files = self.find_tensor_file(name, slot, step - 1, rank_id, root_graph_id)
        tensor_files.sort()
        try:
            file_path = os.path.join(iteration_path, tensor_files[-1])
        except IndexError:
            log.error("Can not find the file: name %s, in iteration %s, rank_id %s", name, step, rank_id)
            raise DebuggerParamValueError("Can not find the file: name {}, in iteration {}, rank_id {}".format(
                name, step, rank_id))

        node_name = name.rsplit('/')[-1]
        file_name = get_download_file_name(load_info.get("node_type"), node_name, slot)
        add_to_download_mgr(file_name, file_path, None, load_info, step, rank_id, self._cache_store)


class CommandListener:
    """Event listener."""

    def __init__(self, cache_store):
        self._cache_store = cache_store
        self._metadata_stream = cache_store.get_stream_handler(Streams.METADATA)
        # the next position of command queue to be queried
        self._pos = '0'
        self._is_waiting = Event()

    def start(self):
        """Start event listener."""
        self._pos = '0'
        self._is_waiting.set()

    def stop(self):
        """Stop event listener."""
        # stop waiting for new user commands but can still get old commands.
        self._is_waiting.clear()

    def has_new_command(self):
        """Check if there is new command in command queue."""
        return self._cache_store.has_command(self._pos)

    def get_next_command(self):
        """Get next command."""
        event = None
        while event is None and self.has_new_command():
            self._pos, event = self._cache_store.get_command(self._pos)
            log.debug("Deal with old %s-th command:\n%s.", self._pos, event)
        if event is None:
            event = self._wait_for_next_command()
        return event

    def _wait_for_next_command(self):
        """
        Wait for next command.

        Returns:
            EventReply, the command event.
        """
        if not self._is_waiting.is_set():
            self._metadata_stream.state = ServerStatus.PENDING.value
            return None
        log.info("Start to wait for command.")
        if self._metadata_stream.state != ServerStatus.WAITING.value:
            self._metadata_stream.state = ServerStatus.WAITING.value
            self._cache_store.put_data(self._metadata_stream.get())
        log.debug("Wait for %s-th command", self._pos)
        event = None
        while event is None and self._is_waiting.is_set():
            self._pos, event = self._cache_store.get_command(self._pos)
        return event


def convert_watchpointhit(watchpointhit):
    """Convert watchpointhit object to dict."""
    parameters = watchpointhit.parameters
    param_list = []
    for param in parameters:
        param_dict = convert_param(param)
        param_list.append(param_dict)
    watchpointhit_dict = {'condition': watchpointhit.condition,
                          'rank_id': watchpointhit.rank_id,
                          'error_code': watchpointhit.error_code,
                          'name': watchpointhit.name,
                          'parameters': param_list,
                          'slot': watchpointhit.slot,
                          'watchpoint_id': watchpointhit.watchpoint_id,
                          'root_graph_id': watchpointhit.root_graph_id
                          }
    return watchpointhit_dict


def _convert_tensor_base(tensor_data):
    """Convert tensor object to dict."""
    tensor_data_dict = {
        'dtype': tensor_data.dtype,
        'shape': tensor_data.shape,
        'data_size': tensor_data.data_size
    }
    return tensor_data_dict


def _convert_tensor_value(tensor_data):
    """Convert tensor object to dict."""
    tensor_data_dict = {'dtype': tensor_data.dtype,
                        'shape': tensor_data.shape,
                        'data_size': tensor_data.data_size,
                        'data_ptr': tensor_data.data_ptr}
    return tensor_data_dict


def convert_param(param):
    """Convert parameter object to dict"""
    param_dict = {'actual_value': param.actual_value,
                  'disabled': param.disabled,
                  'hit': param.hit,
                  'name': param.name,
                  'value': param.value}
    return param_dict



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Define the metadata stream handler."""

from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.common.utils import ServerStatus, DebuggerServerMode
from mindinsight.debugger.stream_handler.base_handler import StreamHandlerBase


class MetadataHandler(StreamHandlerBase):
    """Metadata Handler."""

    def __init__(self):
        self._state = ServerStatus.PENDING
        self._device_name = ""
        self.step = 0
        self._client_ip = ""
        self._cur_node_name = ""
        self._cur_full_name = ""
        self.backend = ""
        self._enable_recheck = False
        self._cur_graph_name = ""
        # If recommendation_confirmed is true, it only means the user has answered yes or no to the question,
        # it does not necessarily mean that the user will use the recommended watch points.
        self._recommendation_confirmed = False
        self._debugger_version = {}
        self._data_version = {}
        # maximum step number among all devices
        self._max_step_num = 0
        self._debugger_type = DebuggerServerMode.ONLINE.value
        self.max_graph_node_size = 0

    @property
    def debugger_type(self):
        """The property of debugger_type."""
        return self._debugger_type

    @debugger_type.setter
    def debugger_type(self, debugger_type):
        """The property of debugger_type."""
        self._debugger_type = debugger_type

    @property
    def device_name(self):
        """The property of device name."""
        return self._device_name

    @property
    def node_name(self):
        """The property of current node name."""
        return self._cur_node_name

    @node_name.setter
    def node_name(self, node_name):
        """The property of current node name."""
        self._cur_node_name = node_name

    @property
    def graph_name(self):
        """The property of current node name."""
        return self._cur_graph_name

    @graph_name.setter
    def graph_name(self, graph_name):
        """The property of current node name."""
        self._cur_graph_name = graph_name if graph_name else ''

    @property
    def full_name(self):
        """The property of current node name."""
        return self._cur_full_name

    @property
    def state(self):
        """The property of state."""
        return self._state.value

    @state.setter
    def state(self, value):
        """
        Set the property of state.

        Args:
            value (str): The new state.
        """
        self._state = ServerStatus(value)

    @property
    def client_ip(self):
        """The property of client ip."""
        return self._client_ip

    @client_ip.setter
    def client_ip(self, value):
        """
        Set the property of client ip.

        Args:
            value (str): The new ip.
        """
        self._client_ip = str(value)

    @property
    def enable_recheck(self):
        """The property of enable_recheck."""
        return self._enable_recheck and self._state == ServerStatus.WAITING

    @enable_recheck.setter
    def enable_recheck(self, value):
        """
        Set the property of enable_recheck.

        Args:
            value (bool): The new ip.
        """
        self._enable_recheck = bool(value)

    @property
    def recommendation_confirmed(self):
        """The property of recommendation_confirmed."""
        return self._recommendation_confirmed

    @recommendation_confirmed.setter
    def recommendation_confirmed(self, value):
        """
        Set the property of recommendation_confirmed.

        Args:
            value (str): The new ip.
        """
        self._recommendation_confirmed = value

    @property
    def debugger_version(self):
        """The property of debugger_version."""
        return self._debugger_version

    @debugger_version.setter
    def debugger_version(self, value):
        """
        Set the property of debugger_version.

        Args:
            value (dict): The  semantic versioning of mindinsight and mindspore,
            format is {'ms': 'x.x.x', 'mi': 'x.x.x'}.
        """
        self._debugger_version = value

    @property
    def data_version(self):
        """The property of data_version."""
        return self._data_version

    @data_version.setter
    def data_version(self, value):
        """
        Set the property of data_version.

        Args:
            value (dict): The semantic versioning of mindinsight and mindspore dump data,
            format is {'state': True, 'ms': 'x.x.x', 'mi': 'x.x.x'}.
        """
        self._data_version = value

    @property
    def max_step_num(self):
        """The property of max_step_num."""
        return self._max_step_num

    @max_step_num.setter
    def max_step_num(self, max_step_num):
        """Set the property of max_step_num."""
        self._max_step_num = max_step_num

    def put(self, value):
        """
        Put value into metadata cache. Called by grpc server.

        Args:
            value (MetadataProto): The Metadata proto message.
        """
        self._device_name = value.device_name.split(':')[0]
        self.step = value.cur_step
        self._cur_full_name = value.cur_node
        self.backend = value.backend if value.backend else "Ascend"
        log.debug("Put metadata into cache at the %d-th step.", self.step)

    def get(self, filter_condition=None):
        """
        Get updated value. Called by main server.

        Args:
            filter_condition (Union[str, list[str]]): The filter property.

        Returns:
            dict, the metadata.
        """
        metadata = {}
        if filter_condition is None:
            metadata = {
                'state': self.state,
                'step': self.step,
                'device_name': self.device_name,
                'pos': '0',
                'ip': self.client_ip,
                'node_name': self.node_name,
                'backend': self.backend,
                'enable_recheck': self.enable_recheck,
                'graph_name': self.graph_name,
                'recommendation_confirmed': self._recommendation_confirmed,
                'debugger_version': self.debugger_version,
                'data_version': self.data_version,
            }
            if self.debugger_type == 'offline':
                metadata['total_step_num'] = self.max_step_num
            if self.state == ServerStatus.NODE_TOO_LARGE.value:
                metadata['max_graph_node_size'] = self.max_graph_node_size
        else:
            if not isinstance(filter_condition, list):
                filter_condition = [filter_condition]
            for field in filter_condition:
                metadata[field] = getattr(self, field) if \
                    hasattr(self, field) else None

        return {'metadata': metadata}



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Define the Device stream handler."""
from collections import defaultdict

from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamValueError, DeviceIdUnregistered, \
    DebuggerParamTypeError
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.stream_handler.base_handler import StreamHandlerBase


class DeviceHandler(StreamHandlerBase):
    """Metadata Handler."""

    def __init__(self):
        # contains all device infos, the format is like Dict[int(<device_id>, <device_info>)]
        self._rank_info = defaultdict(DeviceInfo)
        self._device_rank_map = {}

    @property
    def rank_ids(self):
        """The rank ids."""
        return list(self._rank_info)

    @property
    def device_amount(self):
        """The rank ids."""
        return len(self._rank_info)

    def put(self, value):
        """
        Put value into device info cache.

        Args:
            value (list): The list of server info. Each item is format like:
                {
                    "server_id": str,
                    "device": list[<Device Info>]
                },
                The format of <Device Info> is like:
                {
                    "device_id": str,
                    "device_ip": str,
                    "rank_id": str
                }.
        """
        if not isinstance(value, list):
            log.error("Invalid input type. list object is expected.")
            raise DebuggerParamTypeError("List object is expected.")
        try:
            self._extract_rank_info(value)
        except (TypeError, ValueError) as err:
            log.exception(err)
            log.error("Invalid Device info.")
            raise DebuggerParamValueError("Invalid device info.")
        log.debug("Put Device into cache")

    def _extract_rank_info(self, value):
        """Extract rank info and save."""
        for server_info in value:
            server_ip = server_info.get('server_id')
            for device_info in server_info.get('device', []):
                rank_id = int(device_info.get('rank_id'))
                if rank_id in self._rank_info:
                    log.error("Repeated rank info for rank_id: %d", rank_id)
                    raise DebuggerParamValueError("Repeated rank info.")
                device_info_obj = self._rank_info[rank_id]
                device_info_obj.rank_id = rank_id
                device_info_obj.server_ip = server_ip
                device_info_obj.device_id = int(device_info.get('device_id'))
                device_info_obj.device_ip = device_info.get('device_ip')
                self._device_rank_map[device_info_obj.device_id] = rank_id

    def add_step_num_info(self, step_info):
        """
        Add step number information for each rank.

        Args:
            step_info (dict): Step info per rank. The key is the rank id, the value
                is the relative step number.
        """
        if not step_info:
            log.warning("No step number information.")
            return
        for rank_id, step_num in step_info.items():
            self._rank_info[rank_id].step_num = step_num

    def add_graph_name_info(self, graphs):
        """
        Add graph name per device.

        Args:
            graphs (dict): Graph infos of all rank id. Each item is format like
        """
        for rank_id, graph_info in graphs.items():
            graph_names = list(graph_info)
            if len(graph_names) > 1:
                # if more than one graphs in a device, sort them
                # by the number following the last "_" in the graph_name
                graph_names = sorted(graph_names, key=lambda x: x.split("_")[-1])
            self._rank_info[rank_id].graph_names = graph_names

    def get(self, filter_condition=None):
        """
        Get device information according to filter_condition.

        Args:
            filter_condition (list): The rank id.

        Returns:
            dict, the device info.
        """
        if filter_condition is None:
            filter_condition = self.rank_ids
        if not isinstance(filter_condition, list):
            filter_condition = [filter_condition]
        device_infos = []
        for rank_id in filter_condition:
            device_info = self._rank_info.get(rank_id)
            if device_info is None:
                log.error("Invalid rank id.")
                raise DeviceIdUnregistered(rank_id)
            device_infos.append(device_info.to_dict())
        return {'devices': device_infos}

    def get_rank_id_by_device_id(self, device_id):
        """
        Get rank id by device id.

        Args:
            device_id (int): The device id.

        Returns:
            int, the rank id.
        """
        rank_id = self._device_rank_map.get(device_id)
        if rank_id is None:
            log.error("Failed to find rank_id for device_id %s", device_id)
            raise DeviceIdUnregistered(device_id)
        return rank_id

    def get_device_id_by_rank_id(self, rank_id):
        """
        Get device id by rank id.

        Args:
            rank_id (int): The rank id.

        Returns:
            int, the device id.
        """
        device_info = self._rank_info.get(rank_id)
        if device_info:
            return device_info.device_id
        log.error("Failed to find device id according to rank_id %s", rank_id)
        raise DeviceIdUnregistered(rank_id)


class DeviceInfo:
    """Device info object."""

    def __init__(self):
        self.rank_id = 0
        self.device_id = 0
        self.server_ip = ''
        self.graph_names = []
        self.device_ip = ''
        self.step_num = 0

    def to_dict(self):
        """Convert device info to dict."""
        res = {
            'rank_id': self.rank_id,
            'server_ip': self.server_ip,
            'device_id': self.device_id,
            'device_ip': self.device_ip,
            'graph_names': self.graph_names,
            'total_step_num': self.step_num
        }
        return res



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Definition of error code and relative messages in profiler module."""
from mindinsight.profiler.common.exceptions.error_code import ProfilerErrors, \
    ProfilerErrorMsg
from mindinsight.utils.exceptions import MindInsightException


class ProfilerParamValueErrorException(MindInsightException):
    """The parameter value error in profiler module."""

    def __init__(self, msg):
        super(ProfilerParamValueErrorException, self).__init__(
            error=ProfilerErrors.PARAM_VALUE_ERROR,
            message=ProfilerErrorMsg.PARAM_VALUE_ERROR.value.format(msg),
            http_code=400
        )


class ProfilerPathErrorException(MindInsightException):
    """The path error in profiler module."""

    def __init__(self, msg):
        super(ProfilerPathErrorException, self).__init__(
            error=ProfilerErrors.PATH_ERROR,
            message=ProfilerErrorMsg.PATH_ERROR.value.format(msg),
            http_code=400
        )


class ProfilerParamTypeErrorException(MindInsightException):
    """The parameter type error in profiler module."""

    def __init__(self, msg):
        super(ProfilerParamTypeErrorException, self).__init__(
            error=ProfilerErrors.PARAM_TYPE_ERROR,
            message=ProfilerErrorMsg.PARAM_TYPE_ERROR.value.format(msg),
            http_code=400
        )


class ProfilerDirNotFoundException(MindInsightException):
    """The dir not found exception in profiler module."""

    def __init__(self, msg):
        super(ProfilerDirNotFoundException, self).__init__(
            error=ProfilerErrors.DIR_NOT_FOUND_ERROR,
            message=ProfilerErrorMsg.DIR_NOT_FOUND_ERROR.value.format(msg),
            http_code=400
        )


class ProfilerInfoFieldMissingException(MindInsightException):
    """Some fields in profiler info are missed."""

    def __init__(self, msg):
        super(ProfilerInfoFieldMissingException, self).__init__(
            error=ProfilerErrors.FIELD_MISSED_ERROR,
            message=msg,
            http_code=400
        )


class ProfilerFileNotFoundException(MindInsightException):
    """The file not found exception in profiler module."""

    def __init__(self, msg):
        super(ProfilerFileNotFoundException, self).__init__(
            error=ProfilerErrors.FILE_NOT_FOUND_ERROR,
            message=ProfilerErrorMsg.FILE_NOT_FOUND_ERROR.value.format(msg),
            http_code=400
        )


class ProfilerIOException(MindInsightException):
    """The IO exception in profiler module."""

    def __init__(self):
        super(ProfilerIOException, self).__init__(
            error=ProfilerErrors.IO_ERROR,
            message=ProfilerErrorMsg.IO_ERROR.value,
            http_code=400
        )


class ProfilerDeviceIdMismatchException(MindInsightException):
    """The device id mismatch exception in profiler module."""

    def __init__(self):
        super(ProfilerDeviceIdMismatchException, self).__init__(
            error=ProfilerErrors.DEVICE_ID_MISMATCH_ERROR,
            message=ProfilerErrorMsg.DEVICE_ID_MISMATCH_ERROR.value,
            http_code=400
        )


class ProfilerRawFileException(MindInsightException):
    """The raw file exception in profiler module."""

    def __init__(self, msg):
        super(ProfilerRawFileException, self).__init__(
            error=ProfilerErrors.RAW_FILE_ERROR,
            message=ProfilerErrorMsg.RAW_FILE_ERROR.value.format(msg),
            http_code=400
        )


class ProfilerColumnNotExistException(MindInsightException):
    """The column does not exist exception in profiler module."""

    def __init__(self, msg):
        super(ProfilerColumnNotExistException, self).__init__(
            error=ProfilerErrors.COLUMN_NOT_EXIST_ERROR,
            message=ProfilerErrorMsg.COLUMN_NOT_EXIST_ERROR.value.format(msg),
            http_code=400
        )


class ProfilerAnalyserNotExistException(MindInsightException):
    """The analyser in profiler module."""

    def __init__(self, msg):
        super(ProfilerAnalyserNotExistException, self).__init__(
            error=ProfilerErrors.ANALYSER_NOT_EXIST_ERROR,
            message=ProfilerErrorMsg.ANALYSER_NOT_EXIST_ERROR.value.format(msg),
            http_code=400
        )


class ProfilerDeviceIdException(MindInsightException):
    """The parameter device_id error in profiler module."""

    def __init__(self, msg):
        super(ProfilerDeviceIdException, self).__init__(
            error=ProfilerErrors.DEVICE_ID_ERROR,
            message=ProfilerErrorMsg.DEIVICE_ID_ERROR.value.format(msg),
            http_code=400
        )


class ProfilerOpTypeException(MindInsightException):
    """The parameter op_type error in profiler module."""

    def __init__(self, msg):
        super(ProfilerOpTypeException, self).__init__(
            error=ProfilerErrors.OP_TYPE_ERROR,
            message=ProfilerErrorMsg.OP_TYPE_ERROR.value.format(msg),
            http_code=400
        )


class ProfilerSortConditionException(MindInsightException):
    """The parameter sort_condition error in profiler module."""

    def __init__(self, msg):
        super(ProfilerSortConditionException, self).__init__(
            error=ProfilerErrors.SORT_CONDITION_ERROR,
            message=ProfilerErrorMsg.SORT_CONDITION_ERROR.value.format(msg),
            http_code=400
        )


class ProfilerFilterConditionException(MindInsightException):
    """The parameter filer_condition error in profiler module."""

    def __init__(self, msg):
        super(ProfilerFilterConditionException, self).__init__(
            error=ProfilerErrors.FILTER_CONDITION_ERROR,
            message=ProfilerErrorMsg.FILTER_CONDITION_ERROR.value.format(msg),
            http_code=400
        )


class ProfilerGroupConditionException(MindInsightException):
    """The parameter group_condition error in profiler module."""

    def __init__(self, msg):
        super(ProfilerGroupConditionException, self).__init__(
            error=ProfilerErrors.GROUP_CONDITION_ERROR,
            message=ProfilerErrorMsg.GROUP_CONDITION_ERROR.value.format(msg),
            http_code=400
        )


class ProfilerColumnNotSupportSortException(MindInsightException):
    """The column does not support to sort error in profiler module."""

    def __init__(self, msg):
        super(ProfilerColumnNotSupportSortException, self).__init__(
            error=ProfilerErrors.COLUMN_NOT_SUPPORT_SORT_ERROR,
            message=ProfilerErrorMsg.COLUMN_NOT_SUPPORT_SORT_ERROR.value.format(msg),
            http_code=400
        )


class StepNumNotSupportedException(MindInsightException):
    """The step number error in profiler module."""

    def __init__(self, msg):
        super(StepNumNotSupportedException, self).__init__(
            error=ProfilerErrors.STEP_NUM_NOT_SUPPORTED_ERROR,
            message=ProfilerErrorMsg.STEP_NUM_NOT_SUPPORTED_ERROR.value.format(msg),
            http_code=400
        )


class JobIdMismatchException(MindInsightException):
    """The Job ID mismatch error in profiler module."""

    def __init__(self):
        super(JobIdMismatchException, self).__init__(
            error=ProfilerErrors.JOB_ID_MISMATCH_ERROR,
            message=ProfilerErrorMsg.JOB_ID_MISMATCH_ERROR.value,
            http_code=400
        )


class ProfilerPipelineOpNotExistException(MindInsightException):
    """The minddata pipeline operator does not exist error in profiler module."""

    def __init__(self, msg):
        super(ProfilerPipelineOpNotExistException, self).__init__(
            error=ProfilerErrors.PIPELINE_OP_NOT_EXIST_ERROR,
            message=ProfilerErrorMsg.PIPELINE_OP_NOT_EXIST_ERROR.value.format(msg),
            http_code=400
        )


class UnsupportedParallelTypeException(MindInsightException):
    """The provided parallel type is unsupported."""

    def __init__(self, detail):
        super(UnsupportedParallelTypeException, self).__init__(
            error=ProfilerErrors.UNSUPPORTED_PARALLEL_TYPE_ERROR,
            message=ProfilerErrorMsg.UNSUPPORTED_PARALLEL_TYPE_ERROR.value.format(detail),
            http_code=400
        )


class WrongParallelStrategyDataException(MindInsightException):
    """The provided parallel strategy data is failed."""

    def __init__(self, detail):
        super(WrongParallelStrategyDataException, self).__init__(
            error=ProfilerErrors.WRONG_PARALLEL_STRATEGY_DATA_ERROR,
            message=ProfilerErrorMsg.WRONG_PARALLEL_STRATEGY_DATA_ERROR.value.format(detail),
            http_code=400
        )


class NotFoundParallelStrategyDataException(MindInsightException):
    """Can not find any parallel strategy data."""

    def __init__(self):
        super(NotFoundParallelStrategyDataException, self).__init__(
            error=ProfilerErrors.NOT_FOUND_PARALLEL_STRATEGY_DATA_ERROR,
            message=ProfilerErrorMsg.NOT_FOUND_PARALLEL_STRATEGY_DATA_ERROR.value,
            http_code=400
        )


class FileNumNotMatchException(MindInsightException):
    """Wrong parallel strategy data num."""

    def __init__(self):
        super(FileNumNotMatchException, self).__init__(
            error=ProfilerErrors.WRONG_PARALLEL_STRATEGY_DATA_NUM_ERROR,
            message=ProfilerErrorMsg.WRONG_PARALLEL_STRATEGY_DATA_NUM_ERROR.value,
            http_code=400
        )


class ParseParallelStrategyDataException(MindInsightException):
    """Parse parallel file failed with unknown error."""

    def __init__(self, detail):
        super(ParseParallelStrategyDataException, self).__init__(
            error=ProfilerErrors.PARSE_PARALLEL_STRATEGY_DATA_ERROR,
            message=ProfilerErrorMsg.PARSE_PARALLEL_STRATEGY_DATA_ERROR.value.format(detail),
            http_code=500
        )



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Task manager api.

This module provides the interfaces to task manage functions.
"""

import os

from flask import Blueprint
from flask import request
from flask import jsonify

from mindinsight.conf import settings
from mindinsight.utils.exceptions import ParamMissError
from mindinsight.datavisual.common.validation import Validation
from mindinsight.datavisual.data_transform.summary_watcher import SummaryWatcher
from mindinsight.datavisual.utils.tools import str_to_bool
from mindinsight.datavisual.processors.train_task_manager import TrainTaskManager
from mindinsight.datavisual.data_transform.data_manager import DATA_MANAGER


BLUEPRINT = Blueprint("task_manager", __name__, url_prefix=settings.URL_PATH_PREFIX+settings.API_PREFIX)


@BLUEPRINT.route("/datavisual/single-job", methods=["GET"])
def query_single_train_task():
    """Query single train task"""
    plugin_name = request.args.get('plugin_name')
    train_id = request.args.get('train_id')

    processor = TrainTaskManager(DATA_MANAGER)
    tasks = processor.get_single_train_task(train_id=train_id, plugin_name=plugin_name)
    return jsonify(tasks)


@BLUEPRINT.route("/datavisual/plugins", methods=["GET"])
def query_plugins():
    """Query plugins."""
    train_id = request.args.get('train_id')

    manual_update = request.args.get('manual_update', default='false')
    manual_update = str_to_bool(manual_update, "manual_update")

    processor = TrainTaskManager(DATA_MANAGER)
    plugins = processor.get_plugins(train_id, manual_update)
    return jsonify(plugins)


@BLUEPRINT.route("/datavisual/train-jobs", methods=["GET"])
def query_train_jobs():
    """Query train jobs."""
    offset = request.args.get("offset", default=0)
    limit = request.args.get("limit", default=10)
    train_id = request.args.get('train_id')

    offset = Validation.check_offset(offset=offset)
    limit = Validation.check_limit(limit, min_value=1, max_value=SummaryWatcher.MAX_SUMMARY_DIR_COUNT)

    processor = TrainTaskManager(DATA_MANAGER)
    total, train_jobs, load_status = processor.query_train_jobs(offset, limit, train_id)

    return jsonify({
        'name': os.path.basename(os.path.realpath(settings.SUMMARY_BASE_DIR)),
        'total': total,
        'train_jobs': train_jobs,
        'load_status': load_status
    })


@BLUEPRINT.route("/datavisual/train-job-caches", methods=["POST"])
def cache_train_jobs():
    """ Cache train jobs."""
    data = request.get_json(silent=True)
    if data is None:
        raise ParamMissError('train_ids')

    train_ids = data.get('train_ids')
    if train_ids is None:
        raise ParamMissError('train_ids')

    processor = TrainTaskManager(DATA_MANAGER)
    cache_result = processor.cache_train_jobs(train_ids)

    return jsonify({'cache_result': cache_result})


def init_module(app):
    """
    Init module entry.

    Args:
        app: the application obj.

    """
    app.register_blueprint(BLUEPRINT)



# Copyright 2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Marey api.

This module provides the interfaces to profile used in extended marey's graph.
"""
import os

from flask import Blueprint, request, jsonify
from marshmallow import ValidationError

from mindinsight.conf import settings
from mindinsight.profiler.common.validator.validate_path import validate_and_normalize_path
from mindinsight.profiler.common.exceptions.exceptions import ProfilerFileNotFoundException
from mindinsight.profiler.analyser.analyser_factory import AnalyserFactory
from mindinsight.profiler.common.util import check_train_job_and_profiler_dir
from mindinsight.profiler.common.log import logger
from mindinsight.utils.exceptions import ParamValueError

BLUEPRINT = Blueprint("profile_marey_graph", __name__, url_prefix=settings.URL_PATH_PREFIX + settings.API_PREFIX)


def init_module(app):
    """
    Init module entry.

    Args:
        app: the application obj.

    """
    app.register_blueprint(BLUEPRINT)


@BLUEPRINT.route("/profile/marey-graph/memory-data", methods=["GET"])
def get_memory_data():
    """Get memory data by train id."""

    train_id = request.args.get('train_id')
    device_type = request.args.get("device_type")
    if device_type not in ['ascend']:
        logger.info("Invalid device_type, device_type should be gpu or ascend.")
        raise ParamValueError("Invalid device_type.")
    if train_id is None:
        logger.info("Invalid train_id parameter of None.")
        raise ParamValueError("Invalid train_id.")

    profiler_dir = os.path.realpath(os.path.join(settings.SUMMARY_BASE_DIR, train_id, 'profiler'))
    try:
        profiler_dir = validate_and_normalize_path(profiler_dir, 'profiler')
    except ValidationError as exc:
        raise ProfilerFileNotFoundException('Invalid cluster_profiler dir, detail: %s.' % str(exc))

    check_train_job_and_profiler_dir(profiler_dir)

    analyser = AnalyserFactory.instance().get_analyser('memory_usage', profiler_dir, train_id)
    return jsonify(analyser.get_memory_data_for_marey(device_type))


@BLUEPRINT.route("/profile/marey-graph/flops-data", methods=["GET"])
def get_flops_data():
    """Get memory data by train id."""

    train_id = request.args.get('train_id')
    if train_id is None:
        logger.info("Invalid train_id parameter of None.")
        raise ParamValueError("Invalid train_id.")

    profiler_dir = os.path.realpath(os.path.join(settings.SUMMARY_BASE_DIR, train_id, 'profiler'))
    try:
        profiler_dir = validate_and_normalize_path(profiler_dir, 'profiler')
    except ValidationError as exc:
        raise ProfilerFileNotFoundException('Invalid cluster_profiler dir, detail: %s.' % str(exc))

    check_train_job_and_profiler_dir(profiler_dir)

    analyser = AnalyserFactory.instance().get_analyser('flops', profiler_dir, train_id)
    return jsonify(analyser.get_flops_data_for_marey())


@BLUEPRINT.route("/profile/marey-graph/overview-time", methods=["GET"])
def get_overview_time():
    """Get memory data by train id."""

    train_id = request.args.get('train_id')
    if train_id is None:
        logger.info("Invalid train_id parameter of None.")
        raise ParamValueError("Invalid train_id.")

    profiler_dir = os.path.realpath(os.path.join(settings.SUMMARY_BASE_DIR, train_id, 'profiler'))
    try:
        profiler_dir = validate_and_normalize_path(profiler_dir, 'profiler')
    except ValidationError as exc:
        raise ProfilerFileNotFoundException('Invalid cluster_profiler dir, detail: %s.' % str(exc))

    check_train_job_and_profiler_dir(profiler_dir)

    analyser = AnalyserFactory.instance().get_analyser('cluster_step_trace', profiler_dir, train_id)
    return jsonify(analyser.get_overview_time_info())


@BLUEPRINT.route("/profile/marey-graph/timeline", methods=["GET"])
def get_timeline():
    """Get memory data by train id."""

    train_id = request.args.get('train_id')
    device_list = request.args.get('device_list')
    if train_id is None:
        logger.info("Invalid train_id parameter of None.")
        raise ParamValueError("Invalid train_id.")

    step = request.args.get('step', default='1')
    device_type = request.args.get("device_type")
    if device_type not in ['ascend']:
        logger.info("Invalid device_type, device_type should be gpu or ascend.")
        raise ParamValueError("Invalid device_type.")

    profiler_dir = os.path.realpath(os.path.join(settings.SUMMARY_BASE_DIR, train_id, 'profiler'))
    try:
        profiler_dir = validate_and_normalize_path(profiler_dir, 'profiler')
    except ValidationError as exc:
        raise ProfilerFileNotFoundException('Invalid cluster_profiler dir, detail: %s.' % str(exc))

    check_train_job_and_profiler_dir(profiler_dir)

    analyser = AnalyserFactory.instance().get_analyser('timeline', profiler_dir, train_id)
    return jsonify(analyser.get_marey_timeline(step, device_list))



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Local File System."""
import io
import os

from mindinsight.datavisual.common import exceptions
from mindinsight.datavisual.utils.tools import to_str
from mindinsight.datavisual.data_access.base_file_system import BaseFileSystem
from mindinsight.datavisual.data_access.base_file_system import StatInfo
from mindinsight.utils.exceptions import PathNotExistError


class LocalFileSystem(BaseFileSystem):
    """Local file system."""

    def list_dir(self, path):
        """
        List directories by path.

        Args:
            path (str): Directory path or file path.

        Returns:
           list[str], directories.
        """
        path = to_str(path)
        if not self.is_dir(path):
            raise exceptions.PathNotDirectoryError("Path is %s." % path)
        return os.listdir(path)

    def is_dir(self, path):
        """
        Determine if it is a directory.

        Args:
            path (str): Directory path or file path.

        Returns:
            bool, if it is a directory path, return True.
        """
        return os.path.isdir(to_str(path))

    def is_file(self, path):
        """
        Determine if it is a file.

        Args:
            path (str): Directory path or file path.

        Returns:
            bool, if it is a file path, return True.
        """
        return os.path.isfile(to_str(path))

    def exists(self, path):
        """
        Determine if it exists.

        Args:
            path (str): Directory path or file path.

        Returns:
            bool, if it exists, return True.
        """
        return os.path.exists(to_str(path))

    def file_stat(self, file_path):
        """
        Get file stat information.

        Args:
            file_path (str): File path.

        Returns:
            Nametuple, the (size, mtime) of file.
        """
        try:
            file_info = os.stat(to_str(file_path))
        except OSError:
            raise PathNotExistError("File %s is not exist." % file_path)
        return StatInfo(size=file_info.st_size, mtime=file_info.st_mtime)

    @staticmethod
    def read_access(file_path):
        """
        Determine if it has read permission.

        Args:
            file_path (str): File path.

        Returns:
            bool, if it has read permission, return True.
        """
        return os.access(to_str(file_path), os.R_OK)

    def join(self, path, *paths):
        """
        Join paths.

        Args:
            path (str): Directory path.
            paths (str): Path or paths.

        Returns:
            str, the joined path.
        """
        return os.path.join(path, *paths)

    @staticmethod
    def read(file_path, binary_mode=False, size=None, offset=None):
        """
        Read file.

        Args:
            file_path (str): File path.
            binary_mode (bool): If true, mode will be 'rb'. Else, 'r'.
            size (int): Size of bytes to read.
            offset (int): Offset of file to read.
        Returns:
            bytes, the content read.
        """
        mode = "rb" if binary_mode else "r"
        encoding = None if binary_mode else "utf8"
        with io.open(file_path, mode, encoding=encoding) as file:
            if offset is not None:
                file.seek(offset)
            if size is not None:
                return file.read(size)

            return file.read()



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Datafile encapsulator."""

import io
import os

import numpy as np
from PIL import Image

from mindinsight.datavisual.common.exceptions import ImageNotExistError
from mindinsight.explainer.common.enums import ImageQueryTypes
from mindinsight.explainer.encapsulator._hoc_pil_apply import EditStep, pil_apply_edit_steps
from mindinsight.explainer.encapsulator.explain_data_encap import ExplainDataEncap
from mindinsight.utils.exceptions import FileSystemPermissionError
from mindinsight.utils.exceptions import UnknownError

# Max uint8 value. for converting RGB pixels to [0,1] intensity.
_UINT8_MAX = 255

# Color of low saliency.
_SALIENCY_CMAP_LOW = (55, 25, 86, 255)

# Color of high saliency.
_SALIENCY_CMAP_HI = (255, 255, 0, 255)

# Channel modes.
_SINGLE_CHANNEL_MODE = "L"
_RGBA_MODE = "RGBA"
_RGB_MODE = "RGB"

_PNG_FORMAT = "PNG"


def _clean_train_id_b4_join(train_id):
    """Clean train_id before joining to a path."""
    if train_id.startswith("./") or train_id.startswith(".\\"):
        return train_id[2:]
    return train_id


class DatafileEncap(ExplainDataEncap):
    """Datafile encapsulator."""

    def query_image_binary(self, train_id, image_path, image_type):
        """
        Query image binary content.

        Args:
            train_id (str): Job ID.
            image_path (str): Image path relative to explain job's summary directory.
            image_type (str): Image type, Options: 'original', 'overlay' or 'outcome'.

        Returns:
            bytes, image binary content for UI to demonstrate.
        """
        if image_type == ImageQueryTypes.OUTCOME.value:
            return self._get_hoc_image(image_path, train_id)

        abs_image_path = os.path.join(self.job_manager.summary_base_dir,
                                      _clean_train_id_b4_join(train_id),
                                      image_path)

        if self._is_forbidden(abs_image_path):
            raise FileSystemPermissionError("Forbidden.")

        try:

            if image_type != ImageQueryTypes.OVERLAY.value:
                # no need to convert
                with open(abs_image_path, "rb") as fp:
                    return fp.read()

            image = Image.open(abs_image_path)

            if image.mode == _RGBA_MODE:
                # It is RGBA already, do not convert.
                with open(abs_image_path, "rb") as fp:
                    return fp.read()

        except FileNotFoundError:
            raise ImageNotExistError(f"train_id:{train_id} path:{image_path} type:{image_type}")
        except PermissionError:
            raise FileSystemPermissionError(f"train_id:{train_id} path:{image_path} type:{image_type}")
        except OSError:
            raise UnknownError(f"Invalid image file: train_id:{train_id} path:{image_path} type:{image_type}")

        if image.mode == _SINGLE_CHANNEL_MODE:
            saliency = np.asarray(image) / _UINT8_MAX
        elif image.mode == _RGB_MODE:
            saliency = np.asarray(image)
            saliency = saliency[:, :, 0] / _UINT8_MAX
        else:
            raise UnknownError(f"Invalid overlay image mode:{image.mode}.")

        saliency_stack = np.empty((saliency.shape[0], saliency.shape[1], 4))
        for c in range(3):
            saliency_stack[:, :, c] = saliency
        rgba = saliency_stack * _SALIENCY_CMAP_HI
        rgba += (1 - saliency_stack) * _SALIENCY_CMAP_LOW
        rgba[:, :, 3] = saliency * _UINT8_MAX

        overlay = Image.fromarray(np.uint8(rgba), mode=_RGBA_MODE)
        buffer = io.BytesIO()
        overlay.save(buffer, format=_PNG_FORMAT)

        return buffer.getvalue()

    def _is_forbidden(self, path):
        """Check if the path is outside summary base dir."""
        base_dir = os.path.realpath(self.job_manager.summary_base_dir)
        path = os.path.realpath(path)
        return not path.startswith(base_dir)

    def _get_hoc_image(self, image_path, train_id):
        """Get hoc image for image data demonstration in UI."""

        sample_id, label, layer = image_path.strip(".jpg").split("_")
        layer = int(layer)
        job = self.job_manager.get_job(train_id)
        samples = job.samples
        label_idx = job.labels.index(label)

        chosen_sample = samples[int(sample_id)]
        original_path_image = chosen_sample['image']
        abs_image_path = os.path.join(self.job_manager.summary_base_dir, _clean_train_id_b4_join(train_id),
                                      original_path_image)
        if self._is_forbidden(abs_image_path):
            raise FileSystemPermissionError("Forbidden.")

        image_type = ImageQueryTypes.OUTCOME.value
        try:
            image = Image.open(abs_image_path)
        except FileNotFoundError:
            raise ImageNotExistError(f"train_id:{train_id} path:{image_path} type:{image_type}")
        except PermissionError:
            raise FileSystemPermissionError(f"train_id:{train_id} path:{image_path} type:{image_type}")
        except OSError:
            raise UnknownError(f"Invalid image file: train_id:{train_id} path:{image_path} type:{image_type}")

        edit_steps = []
        boxes = chosen_sample["hierarchical_occlusion"][label_idx]["hoc_layers"][layer]["boxes"]
        mask = chosen_sample["hierarchical_occlusion"][label_idx]["mask"]

        for box in boxes:
            edit_steps.append(EditStep(layer, *box))
        image_cp = pil_apply_edit_steps(image, mask, edit_steps)
        buffer = io.BytesIO()
        image_cp.save(buffer, format=_PNG_FORMAT)

        return buffer.getvalue()
